"use client";

import "./AppHeader.scss";
import {
	Button,
	ButtonGroup,
	Header,
	HeaderLogo,
	IconButton,
	Input,
	Text,
	Tooltip,
} from "@itwin/itwinui-react";
import { SvgMoon, SvgSun } from "@itwin/itwinui-icons-react";
import {
	AppThemeContext,
	MatchesContext,
	RepoFolderContext,
	RulesContext,
} from "modules/contexts/contexts";
import React from "react";
import { displayToast, getAllMatchesInRepo } from "utils/tasks";
import { removeMultipleSlashes } from "utils/utils";
import path from "path";
import Link from "next/link";
import { usePathname } from "next/navigation";
import { getOperatingSystemType } from "utils/utils";
import cx from "classnames";

export const AppHeader = (props: {
	showInput?: boolean;
	className?: string;
}) => {
	const { className, showInput = true } = props;

	const pathname = usePathname();
	const headerLogoLink = pathname === "/" ? null : "/";
	const showLogoText = pathname !== "/";

	const { packagesToScan } = React.useContext(RulesContext);
	const { repoFolder, setFoundRepoFolder, setFolderInfo } = React.useContext(RepoFolderContext);
	const { theme, setTheme } = React.useContext(AppThemeContext);
	const { setMatches, gettingMatches, setGettingMatches, setFileTree } =
		React.useContext(MatchesContext);

	const [repoFolderTextInput, setRepoFolderTextInput] = React.useState<string>("");

	const { repoFolderExample } = useExampleRepoFolderExample();

	const repoFolderInputForm = React.useCallback(
		() => (
			<form
				onSubmit={async (e) => {
					e.preventDefault();

					if (packagesToScan.length === 0) {
						displayToast("negative", "Choose at least one rule set");
						return;
					}

					const cleanedRepoFolderTextInput = removeMultipleSlashes(repoFolderTextInput, window);
					setRepoFolderTextInput(path.normalize(cleanedRepoFolderTextInput));

					if (repoFolder !== null) {
						repoFolder.current = cleanedRepoFolderTextInput;
					}
					getAllMatchesInRepo(
						repoFolder?.current ?? "",
						setFoundRepoFolder,
						setMatches,
						setFileTree,
						setFolderInfo,
						gettingMatches,
						setGettingMatches,
						packagesToScan,
					);
				}}
			>
				<ButtonGroup className="input-bar" key={1}>
					<Input
						className="input"
						value={repoFolderTextInput}
						onChange={(e) => setRepoFolderTextInput(e.target.value)}
						placeholder={`e.g. ${repoFolderExample}`}
					/>
					<Button className="button" type="submit">
						Scan
					</Button>
				</ButtonGroup>
			</form>
		),
		[
			gettingMatches,
			packagesToScan,
			repoFolder,
			repoFolderExample,
			repoFolderTextInput,
			setFileTree,
			setFolderInfo,
			setFoundRepoFolder,
			setGettingMatches,
			setMatches,
		],
	);

	return (
		<Header
			className={cx("AppHeader", className)}
			appLogo={
				<HeaderLogo
					as={headerLogoLink != null ? Link : "div"}
					href={headerLogoLink != null ? headerLogoLink : ""}
					logo={
						<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16">
							<path d="M15.1 4.73v6.73c0 .48-.27.94-.69 1.18l-5.43 3.2a1.16 1.16 0 0 1-1.17 0c-.37-.22-.59-.6-.59-1.03V9.37a.14.14 0 0 0-.06-.11l-.93-.54a.79.79 0 0 1 .4-1.47.8.8 0 0 1 .4.11l1.08.66c.41.24.67.69.67 1.17v4.96l4.7-2.79c.04-.02.06-.06.06-.1V4.9a.14.14 0 0 0-.07-.11L8.05 1.64a.11.11 0 0 0-.07-.02c-.02 0-.05 0-.06.02l-5.4 3.18a.12.12 0 0 0-.07.1l.01 6.36c0 .04.03.08.06.1l.15.1 2.84 1.65.1.06c.05.05.1.1.14.17a.8.8 0 0 1-.63 1.26c-.1 0-.2-.02-.29-.06l-2.75-1.62-.34-.2-.15-.09a1.37 1.37 0 0 1-.69-1.2V4.74c0-.5.27-.96.7-1.2L7.32.18a1.34 1.34 0 0 1 1.36 0l5.73 3.35c.42.25.68.7.68 1.19ZM8 6.46c.56 0 1.02-.47 1.02-1.04 0-.57-.46-1.03-1.02-1.03s-1.02.46-1.02 1.03c0 .57.46 1.04 1.02 1.04Z" />
						</svg>
					}
				>
					{showLogoText ? (
						<Text key={"title"} className="title" variant="leading">
							iTwinUI Migration Tool
						</Text>
					) : null}
				</HeaderLogo>
			}
			actions={[
				<Tooltip
					key={"icon-button"}
					content={`Change to ${theme === "light" ? "dark" : "light"} theme`}
				>
					<IconButton
						styleType="borderless"
						onClick={() => setTheme((prev: string) => (prev === "light" ? "dark" : "light"))}
					>
						{theme === "light" ? <SvgMoon /> : <SvgSun />}
					</IconButton>
				</Tooltip>,
			]}
		>
			{showInput ? repoFolderInputForm() : null}
		</Header>
	);
};

// ------------------------------------------------------------------------------------------------

/**
 * Repo folder example depending on the OS
 */
export const useExampleRepoFolderExample = () => {
	const exampleWindowsRepoFolder = "C:\\Users\\abc\\Documents\\ProjectRepo";
	const exampleUnixRepoFolder = "/home/abc/Documents/ProjectRepo";
	const [repoFolderExample, setRepoFolderExample] = React.useState(exampleWindowsRepoFolder);

	React.useEffect(() => {
		if (typeof window !== "undefined") {
			setRepoFolderExample(
				getOperatingSystemType(window) === "windows"
					? exampleWindowsRepoFolder
					: exampleUnixRepoFolder,
			);
		}
	}, []);

	return { repoFolderExample };
};
