"use client";

import React from "react";
import { Breadcrumbs } from "@itwin/itwinui-react";
import Link from "next/link";
import styles from "./CustomBreadcrumbs.module.scss";

type SupportedRoutes = "/" | "/v3" | "/v3/preview" | "/v3/results";
const supportedRoutes: Record<
	SupportedRoutes,
	{
		name: string;
		allRoutesInLine?: SupportedRoutes[]; // if not provided, it will be calculated
	}
> = {
	"/": {
		name: "Version Selector",
	},
	"/v3": {
		name: "v2 to v3",
	},
	"/v3/preview": {
		allRoutesInLine: ["/v3", "/v3/preview", "/v3/results"],
		name: "Preview",
	},
	"/v3/results": {
		allRoutesInLine: ["/v3", "/v3/preview", "/v3/results"],
		name: "Results/Suggestions",
	},
};

/**
 * Returns an array of all routes in the route line.
 *
 * @example
 * getAllRoutesInLine({route: "/v3"}) // ["/", "/v3"]
 * getAllRoutesInLine({route: "/v3/suggestions"}) // ["/", "/v3", "/v3/suggestions"]
 */
const getAllRoutesInLine = ({
	route,
}: {
	route: keyof typeof supportedRoutes;
}) => {
	const routeObj = supportedRoutes[route];

	if (routeObj?.allRoutesInLine != null) {
		return routeObj.allRoutesInLine;
	}

	const routeWithoutLeadingSlash = route.substring(1); // remove leading slash
	const routePieces = routeWithoutLeadingSlash.split("/");

	// Handle all routes except root route (We don't want root route in the breadcrumbs)
	const allRoutesInLine = routePieces.reduce((acc, route) => {
		acc.push(`${acc[acc.length - 1] ?? ""}/${route}` as keyof typeof supportedRoutes);
		return acc;
	}, [] as (keyof typeof supportedRoutes)[]);

	return allRoutesInLine;
};

export const CustomBreadcrumbs = ({
	route,
	applyDefaultStyles = true,
}: {
	route: keyof typeof supportedRoutes;
	/**
	 * Whether to apply default styles to the breadcrumbs (e.g. margin-bottom).
	 */
	applyDefaultStyles?: boolean;
}) => {
	const allRoutesInLine = getAllRoutesInLine({ route });

	return (
		<Breadcrumbs className={styles["root"]} data-apply-default-styles={applyDefaultStyles}>
			{allRoutesInLine.map((routeInLine, index) => {
				const isLastRoute = index === allRoutesInLine.length - 1;

				return route !== routeInLine ? (
					// if the route is not the last one, show a <Breadcrumbs.Item> as a Link (href={route} as={Link})
					<Breadcrumbs.Item href={routeInLine} as={Link} key={routeInLine}>
						{supportedRoutes[routeInLine].name}
					</Breadcrumbs.Item>
				) : (
					// if the route is the last one, show a regular <Breadcrumbs.Item>
					<Breadcrumbs.Item key={routeInLine}>{supportedRoutes[routeInLine].name}</Breadcrumbs.Item>
				);
			})}
		</Breadcrumbs>
	);
};
