import { redirect } from "next/navigation";
import { revalidatePath } from "next/cache";
import { runCodemods } from "modules/v3/logic/codemods";
import { writeLatestRunFile } from "utils/latestFileRun";
import { createEmptySuggestionsFile } from "utils/codemods";
import { deleteLatestRunFile } from "utils/latestFileRun";
import { SubmitButton } from "../../_ui/SubmitButton";
import { readOutputFile } from "app/_actions/readOutputFile";
import { PreviousRunCard } from "./PreviousRunCard";
import { PageHeader } from "./PageHeader";
import { LabeledInput } from "@itwin/itwinui-react";
import styles from "./page.module.scss";

export default async function App() {
	const latestRunFile = await readOutputFile();

	const description = `This tool is a set of codemods to help you migrate from v2 to v3 ([migration guide](https://github.com/iTwin/iTwinUI/wiki/iTwinUI-react-v3-migration-guide)). We hope to make as many automatic changes as possible for all the [required (breaking) changes](https://github.com/iTwin/iTwinUI/wiki/iTwinUI-react-v3-migration-guide#breaking-changes). You can see them on the \`Preview page\`. However, there might also be a few manual changes that you need to make. You can see them on the \`Results page\` (available from the \`Preview page\`)
          
  If you ran this previously:
  * We save your previous run. So you can either continue your previous run. Or you can start a new run (would **discard** the old run results).
  
  When starting fresh:
  * Enter the path to your repository below.
    * Only \`.tsx\`, \`.ts\`, \`.js\`, \`.jsx\` files are considered
    * gitignored and untracked files will be ignored
  * Clicking the \`Dry-run codemods\` button will do a dry run of the codemods.
    * **NOTE**: This may take a while, depending on the size of the repo.
  `.trim();

	return (
		<>
			<PageHeader
				customBreadcrumbsProps={{
					route: "/v3",
				}}
				instructionsProps={{
					title: "Home (v2 → v3)",
					description: description,
				}}
			/>
			{latestRunFile == null ? (
				<RepoPathForm />
			) : (
				<PreviousRunCard latestRunFile={latestRunFile} deleteButton={<StartNewRunButton />} />
			)}
		</>
	);
}

// ------------------------------------------------------------------------------------------------

const RepoPathForm = () => {
	const isWindows = process.platform === "win32";
	const repoFolderExample = isWindows
		? "C:\\Users\\abc\\Documents\\ProjectRepo"
		: "/home/abc/Documents/ProjectRepo";

	return (
		<form
			className={styles.repoForm}
			action={async (formData) => {
				"use server";

				let error;

				try {
					const repoFolder = formData.get("repo-path");
					if (typeof repoFolder !== "string") throw "oops";
					await runCodemodsAndWriteToFile(repoFolder);
				} catch (err) {
					console.error("ERROR: ", err);
					error = err;
				}

				if (!error) {
					redirect("/v3/preview");
				}
			}}
		>
			<LabeledInput
				label="Repo path"
				name="repo-path"
				placeholder={`e.g. ${repoFolderExample}`}
				required
				autoFocus
			/>
			<SubmitButton>Dry-run codemods</SubmitButton>
		</form>
	);
};

// ------------------------------------------------------------------------------------------------

async function runCodemodsAndWriteToFile(repoFolder: string) {
	createEmptySuggestionsFile();

	const latestFileRun = await runCodemods({ repoFolder });
	if (latestFileRun == null) throw "oops";

	writeLatestRunFile(latestFileRun);
}

// ------------------------------------------------------------------------------------------------

const StartNewRunButton = () => {
	return (
		<form
			action={async () => {
				"use server";
				deleteLatestRunFile();
				revalidatePath("/v3");
			}}
		>
			<SubmitButton>Start new run</SubmitButton>
		</form>
	);
};
