import {
	Flex,
	InputGroup,
	Checkbox,
	Text,
	Button,
	useToaster,
	Tooltip,
} from "@itwin/itwinui-react";
import type { LatestRunFile } from "utils/latestFileRun";
import { SupportedCodemods } from "codemods/codemods";
import React from "react";
import { PreviewPageContext } from "./PreviewPageContent";
import { bulkUpdateCodemoddedFiles } from "app/_actions/bulkUpdateCodemoddedFiles";
import { PreviewPageOptimisticData } from "./PreviewPageContent";
import { CustomReactMarkdown } from "app/_ui/CustomReactMarkdown";
import { flushSync } from "react-dom";
import { usePersistentToast } from "app/_ui/usePersistentToast";

export const CodemodOptions = ({
	file,
	selectedCodemods,
	setCodemodCheckboxValue,
	optimisticData,
	updateDataOptimistically,
}: {
	file: LatestRunFile["files"][number];
	selectedCodemods: Record<SupportedCodemods, boolean>;
	setCodemodCheckboxValue: (
		codemodToToggle: SupportedCodemods | "all",
		value: boolean,
	) => Promise<void>;
	optimisticData: LatestRunFile;
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
}) => {
	const previewPageContext = React.useContext(PreviewPageContext);
	const isCodemodRunning = previewPageContext?.runningCodemodsSet.has(file.filePath);
	const isAnyFileCodemodsRunning = (previewPageContext?.runningCodemodsSet.size ?? 0) > 0;

	return (
		<Flex flexDirection="column" alignItems="flex-start" gap="s">
			<Flex flexDirection="column" alignItems="flex-start" gap="3xs">
				<Text variant="subheading">Applicable codemods</Text>
				<Text>
					{`Below are all codemods whose activation condition has been triggered.
          Optionally, filter through the codemods for fine-tuning.`}
				</Text>
			</Flex>

			<Flex flexDirection="column" alignItems="flex-start" gap="3xs">
				<FinishedCodemodsList file={file} />
				<InputGroup>
					{/* Select all */}
					<SelectAllCheckbox
						selectedCodemods={selectedCodemods}
						setCodemodCheckboxValue={setCodemodCheckboxValue}
						file={file}
						isCodemodRunning={isCodemodRunning}
					/>

					{/* Codemod options */}
					{file.applicableCodemods.map((codemod) => (
						<CodemodOption
							key={codemod}
							codemod={codemod}
							selectedCodemods={selectedCodemods}
							setCodemodCheckboxValue={setCodemodCheckboxValue}
							file={file}
						/>
					))}
				</InputGroup>
			</Flex>

			{/* Buttons bar */}
			<Flex gap="xs">
				<RunCodemodsOnFileButton
					file={file}
					updateDataOptimistically={updateDataOptimistically}
					isAnyFileCodemodsRunning={isAnyFileCodemodsRunning}
				/>
				<IgnoreFileButton file={file} updateDataOptimistically={updateDataOptimistically} />
			</Flex>
		</Flex>
	);
};

const FinishedCodemodsList = ({
	file,
}: {
	file: LatestRunFile["files"][number];
}) => {
	const content = (
		<>
			<Text as="strong">Applicable codemods:</Text>
			<CustomReactMarkdown>
				{file.applicableCodemods
					.map((applicableCodemod) => {
						const didRun = file.stages.some((stage) => stage.codemod === applicableCodemod);
						return `- ${applicableCodemod} **(${didRun ? "RAN" : "IGNORED"})**`;
					})
					.join("\n")}
			</CustomReactMarkdown>
		</>
	);

	return file.state === "finished-running" ? content : undefined;
};

const SelectAllCheckbox = ({
	selectedCodemods,
	setCodemodCheckboxValue,
	file,
	isCodemodRunning,
}: {
	selectedCodemods: Record<SupportedCodemods, boolean>;
	setCodemodCheckboxValue: (
		codemodToToggle: SupportedCodemods | "all",
		value: boolean,
	) => Promise<void>;
	file: LatestRunFile["files"][number];
	isCodemodRunning?: boolean;
}) => {
	const content = (
		<Checkbox
			onChange={(event) => setCodemodCheckboxValue("all", event.target.checked)}
			label="Select all"
			indeterminate={
				Object.values(file.applicableCodemods).some((selected) => selected) &&
				Object.values(file.applicableCodemods).some((selected) => !selected)
			}
			checked={Object.values(selectedCodemods).every((selected) => selected)}
			disabled={isCodemodRunning || file.state !== "active"}
		/>
	);

	return file.state !== "finished-running" ? content : undefined;
};

const RunCodemodsOnFileButton = ({
	file,
	updateDataOptimistically,
	isAnyFileCodemodsRunning,
}: {
	file: LatestRunFile["files"][number];
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
	isAnyFileCodemodsRunning: boolean | undefined;
}) => {
	const toaster = useToaster();

	const persistentToast = usePersistentToast();

	const previewPageContext = React.useContext(PreviewPageContext);

	const isButtonDisabled = isAnyFileCodemodsRunning;

	const action = async () => {
		if (isAnyFileCodemodsRunning) {
			return;
		}

		updateDataOptimistically({
			action: "run-codemods-on-file",
			filePath: file.filePath,
		});

		flushSync(() => {
			previewPageContext?.setRunningCodemodsSet((prev) => {
				const newSet = new Set(prev);
				newSet.add(file.filePath);
				return newSet;
			});
		});

		const formData = new FormData();
		formData.append("file-path", file.filePath);
		formData.append("new-state", "finished-running");

		persistentToast.start(`Queued codemods to run on file…`);

		const result = await bulkUpdateCodemoddedFiles(formData);

		persistentToast.finish();

		previewPageContext?.setRunningCodemodsSet((prev) => {
			const newSet = new Set(prev);
			newSet.delete(file.filePath);
			return newSet;
		});

		if (result) {
			toaster.positive(
				`Successfully ran ${file.stages.length} ${
					file.stages.length !== 1 ? "codemods" : "codemod"
				} in file`,
			);
		} else {
			toaster.negative("Failed to run codemods in files");
		}
	};

	const button = (
		<Button type="submit" styleType="high-visibility" disabled={isButtonDisabled}>
			{file.state === "active"
				? `Run ${file.stages.length} ${file.stages.length !== 1 ? "codemods" : "codemod"}`
				: file.state === "finished-running"
				? `Finished running codemods`
				: file.state === "ignored"
				? `Ignored`
				: undefined}
		</Button>
	);

	const content = isButtonDisabled ? (
		<Tooltip content="Please wait for any queued runs to finish...">{button}</Tooltip>
	) : (
		button
	);

	return file.state === "active" ? <form action={action}>{content}</form> : undefined;
};

const IgnoreFileButton = ({
	file,
	updateDataOptimistically,
	isCodemodRunning,
}: {
	file: LatestRunFile["files"][number];
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
	isCodemodRunning?: boolean;
}) => {
	const toaster = useToaster();
	const previewPageContext = React.useContext(PreviewPageContext);

	const action = async () => {
		if (isCodemodRunning) {
			return;
		}

		updateDataOptimistically({
			action: file.state === "active" ? "ignore-file" : "un-ignore-file",
			filePath: file.filePath,
		});

		previewPageContext?.setRunningCodemodsSet((prev) => {
			const newSet = new Set(prev);
			newSet.add(file.filePath);
			return newSet;
		});

		const formData = new FormData();
		formData.append("file-path", file.filePath);
		formData.append("new-state", file.state === "ignored" ? "active" : "ignored");

		const newLatestRunFile = await bulkUpdateCodemoddedFiles(formData);
		if (newLatestRunFile == null) {
			toaster.negative("Failed to ignore/un-ignore file");
		}

		previewPageContext?.setRunningCodemodsSet((prev) => {
			const newSet = new Set(prev);
			newSet.delete(file.filePath);
			return newSet;
		});
	};

	const content = (
		<Button type="submit" disabled={isCodemodRunning || file.state === "finished-running"}>
			{file.state === "active"
				? `Ignore file`
				: file.state === "ignored"
				? `Un-ignore file`
				: undefined}
		</Button>
	);

	return file.state !== "finished-running" ? <form action={action}>{content}</form> : undefined;
};

const CodemodOption = ({
	codemod,
	selectedCodemods,
	setCodemodCheckboxValue,
	file,
	isCodemodRunning,
}: {
	codemod: SupportedCodemods;
	selectedCodemods: Record<SupportedCodemods, boolean>;
	setCodemodCheckboxValue: (
		codemodToToggle: SupportedCodemods | "all",
		value: boolean,
	) => Promise<void>;
	file: LatestRunFile["files"][number];
	isCodemodRunning?: boolean;
}) => {
	const content = (
		<Checkbox
			name="codemods-to-run"
			value={codemod}
			onChange={(event) => setCodemodCheckboxValue(codemod, event.target.checked)}
			label={codemod}
			checked={selectedCodemods[codemod]}
			disabled={isCodemodRunning || file.state !== "active"}
		/>
	);

	return file.state !== "finished-running" ? content : undefined;
};
