import {
	SelectionColumn,
	ExpanderColumn,
	Table,
	TablePaginatorRendererProps,
	TablePaginator,
} from "@itwin/itwinui-react";
import { CellProps, Column, TableState, actions, Row } from "@itwin/itwinui-react/react-table";
import type { LatestRunFile } from "utils/latestFileRun";
import React from "react";
import { getFileNameAndPath } from "utils/utils";
import { PreviewPageContext } from "./PreviewPageContent";
import { PreviewPageOptimisticData, previewPageTabs } from "./PreviewPageContent";
import { TableExpandableSubComponent } from "./TableExpandableSubComponent";
import { OpenFileInIDELink } from "app/_ui/OpenFileInIDELink";

type TableProps<T extends Record<string, unknown>> = React.ComponentProps<typeof Table<T>>;

type TableData = {
	filePath: string;
	name: string;
	relativePath: string;
};

export const CodemoddedFilesTable = ({
	activeTab,
	pendingCodemoddedFileCardsToShow,
	finishedRunningCodemoddedFileCardsToShow,
	ignoredCodemoddedFileCardsToShow,
	optimisticData,
	updateDataOptimistically,
}: {
	activeTab: typeof previewPageTabs[number];
	pendingCodemoddedFileCardsToShow: LatestRunFile["files"];
	finishedRunningCodemoddedFileCardsToShow: LatestRunFile["files"];
	ignoredCodemoddedFileCardsToShow: LatestRunFile["files"];
	optimisticData: LatestRunFile;
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
}) => {
	const previewPageContext = React.useContext(PreviewPageContext);

	const isCheckboxDisabled = React.useCallback(
		(rowData: TableData) => {
			return previewPageContext?.runningCodemodsSet.has(rowData.filePath) || false;
		},
		[previewPageContext?.runningCodemodsSet],
	);

	const data = React.useMemo(() => {
		return (
			activeTab === "Pending"
				? pendingCodemoddedFileCardsToShow
				: activeTab === "Finished"
				? finishedRunningCodemoddedFileCardsToShow
				: activeTab === "Ignored"
				? ignoredCodemoddedFileCardsToShow
				: []
		).map((file) => {
			const fileNameAndPath = getFileNameAndPath(file.filePath, optimisticData.repoFolder);

			return {
				filePath: file.filePath,
				name: fileNameAndPath.fileName,
				relativePath: fileNameAndPath.path,
			};
		});
	}, [
		activeTab,
		pendingCodemoddedFileCardsToShow,
		finishedRunningCodemoddedFileCardsToShow,
		ignoredCodemoddedFileCardsToShow,
		optimisticData.repoFolder,
	]);

	const columns = React.useMemo(() => {
		return [
			...(activeTab === "Pending" ? [SelectionColumn({ isDisabled: isCheckboxDisabled })] : []),
			ExpanderColumn({ subComponent: () => true }),
			{
				id: "name",
				Header: "Name",
				accessor: "name",
				Cell: (props: CellProps<TableData>) => {
					return (
						<OpenFileInIDELink filePath={props.row.original.filePath}>
							{props.row.original.name}
						</OpenFileInIDELink>
					);
				},
			},
			{
				id: "relativePath",
				Header: "Relative Path",
				accessor: "relativePath",
			},
		] satisfies Column<TableData>[];
	}, [activeTab, isCheckboxDisabled]);

	const useControlledState: TableProps<TableData>["useControlledState"] = (state) => {
		const selectedRowIds: Record<string, boolean> = {};
		for (const file of previewPageContext?.filteredSelectedRows ?? []) {
			selectedRowIds[file] = true;
		}

		return {
			...state,
			selectedRowIds,
		};
	};

	const stateReducer = React.useCallback(
		(newState: TableState<TableData>, action: typeof actions, prevState: TableState) => {
			if (action.type === "toggleRowSelected" || action.type === "toggleAllRowsSelected") {
				queueMicrotask(() => {
					previewPageContext?.setSelectedRows(
						Object.entries(newState.selectedRowIds)
							.filter(([, selected]) => selected)
							.map(([rowId]) => rowId),
					);
				});
			}

			return newState;
		},
		[previewPageContext],
	);

	const isRowDisabled = React.useCallback(
		(rowData: TableData) => {
			return previewPageContext?.runningCodemodsSet.has(rowData.filePath) ?? false;
		},
		[previewPageContext?.runningCodemodsSet],
	);

	const rowProps = React.useCallback(
		(row: Row<TableData>) => {
			return {
				isLoading: previewPageContext?.runningCodemodsSet.has(row.original.filePath),
			};
		},
		[previewPageContext?.runningCodemodsSet],
	);

	const initialSelectedRows = React.useMemo(() => {
		return optimisticData.files.filter((f) => f.state === "active").map((f) => f.filePath) ?? [];
	}, [optimisticData.files]);

	const initialState = React.useMemo(() => {
		const initialState: Record<string, boolean> = {};
		for (const file of initialSelectedRows) {
			initialState[file] = true;
		}

		return initialState;
	}, [initialSelectedRows]);

	// So that row selection initialization happens only once
	const isAllRowsSelectedInitialized = React.useRef(false);

	// Select all active rows by default if it has at least one change
	React.useEffect(() => {
		if (isAllRowsSelectedInitialized.current || previewPageContext == null) {
			return;
		}

		isAllRowsSelectedInitialized.current = true;

		previewPageContext.setSelectedRows(initialSelectedRows);
	}, [initialSelectedRows, previewPageContext]);

	const isSelectable = activeTab === "Pending";

	return (
		<Table
			emptyTableContent="No codemodded files to show"
			selectRowOnClick={false}
			columns={columns}
			data={data}
			paginatorRenderer={(props: TablePaginatorRendererProps) => (
				<TablePaginator {...props} pageSizeList={[25, 50, 75, 100]} />
			)}
			enableVirtualization={false}
			isSelectable={isSelectable}
			getRowId={(row) => row.filePath}
			useControlledState={isSelectable ? useControlledState : undefined}
			initialState={initialState}
			subComponent={(row) => (
				<TableExpandableSubComponent
					{...row}
					optimisticData={optimisticData}
					updateDataOptimistically={updateDataOptimistically}
				/>
			)}
			stateReducer={isSelectable ? stateReducer : undefined}
			autoResetExpanded={false}
			autoResetSelectedRows={false}
			autoResetFilters={false}
			autoResetPage={false}
			isRowDisabled={isRowDisabled}
			rowProps={rowProps}
		/>
	);
};
