"use client";

import styles from "./PreviewPageContent.module.scss";
import * as React from "react";
import { Flex, Tab, Tabs } from "@itwin/itwinui-react";
import cx from "classnames";
import { RunCodemodsInAllFilesButton } from "./RunCodemodsInAllFilesButton";
import { CodemoddedFilesTable } from "./CodemoddedFilesTable";
import type { LatestRunFile } from "utils/latestFileRun";

// ------------------------------------------------------------------------------------------------

export type PreviewPageOptimisticData = {
	data: LatestRunFile;
	action: {
		action:
			| "run-codemods-on-file"
			| "ignore-file"
			| "un-ignore-file"
			// Temporary dummy action to make `revalidatePath` work
			// See: https://github.com/iTwin/iTwinUI-migration-tool/issues/72
			| "do-nothing";
		filePath: string;
	};
};

const previewPageOptimisticReducer = (
	oldData: LatestRunFile,
	{ action, filePath }: PreviewPageOptimisticData["action"],
) => {
	try {
		if (action === "do-nothing") {
			return oldData;
		}

		const newData = { ...oldData };

		const foundFile = newData.files.find((file) => file.filePath === filePath);
		if (!foundFile) {
			throw null;
		}

		foundFile.state =
			action === "run-codemods-on-file"
				? "finished-running"
				: action === "ignore-file"
				? "ignored"
				: "active";

		return newData;
	} catch (err) {
		return oldData;
	}
};

// ------------------------------------------------------------------------------------------------

export const previewPageTabs = ["Pending", "Finished", "Ignored"] as const;

export const PreviewPageContent = ({
	className,
	latestRunFile,
}: {
	className?: string;
	latestRunFile: LatestRunFile;
}) => {
	const [optimisticData, updateDataOptimistically] = React.useOptimistic(
		latestRunFile,
		previewPageOptimisticReducer,
	);

	const [selectedRows, setSelectedRows] = React.useState<string[]>([]);
	const filteredSelectedRows = selectedRows.filter((file) => {
		const foundFile = optimisticData?.files.find((f) => file === f.filePath);
		return foundFile?.state === "active" && foundFile?.applicableCodemods.length > 0;
	});

	const [runningCodemodsSet, setRunningCodemodsSet] = React.useState<Set<string>>(
		new Set<string>(),
	);

	const pendingCodemoddedFileCardsToShow =
		optimisticData?.files.filter(
			(f) =>
				// Show only active files (codemods haven't been run on them yet)
				f.state === "active",
		) ?? [];

	const finishedRunningCodemoddedFileCardsToShow =
		optimisticData?.files.filter(
			(f) => f.applicableCodemods.length > 0 && f.state === "finished-running",
		) ?? [];

	const ignoredCodemoddedFileCardsToShow =
		optimisticData?.files.filter((f) => f.applicableCodemods.length > 0 && f.state === "ignored") ??
		[];

	return (
		<PreviewPageContext.Provider
			value={{
				selectedRows,
				setSelectedRows,
				filteredSelectedRows,
				runningCodemodsSet,
				setRunningCodemodsSet,
			}}
		>
			<Tabs.Wrapper type="borderless" className={cx(styles.tabsWrapper, className)}>
				<Flex className={styles.tabsListWrapper} justifyContent="space-between">
					<Tabs.TabList className={styles.tabsList}>
						{previewPageTabs.map((tab) => {
							const tabRowsCount =
								{
									Pending: pendingCodemoddedFileCardsToShow?.length,
									Finished: finishedRunningCodemoddedFileCardsToShow?.length,
									Ignored: ignoredCodemoddedFileCardsToShow?.length,
								}[tab] ?? 0;

							return (
								<Tab
									key={`${tab}-${tabRowsCount}`}
									value={tab}
									label={`${tab} (${tabRowsCount})`}
								/>
							);
						})}
					</Tabs.TabList>
					<RunCodemodsInAllFilesButton updateDataOptimistically={updateDataOptimistically} />
				</Flex>

				{previewPageTabs.map((tab) => (
					<Tabs.Panel key={tab} value={tab} className={styles.tabPanel}>
						<CodemoddedFilesTable
							activeTab={tab}
							pendingCodemoddedFileCardsToShow={pendingCodemoddedFileCardsToShow}
							finishedRunningCodemoddedFileCardsToShow={finishedRunningCodemoddedFileCardsToShow}
							ignoredCodemoddedFileCardsToShow={ignoredCodemoddedFileCardsToShow}
							optimisticData={optimisticData}
							updateDataOptimistically={updateDataOptimistically}
						/>
					</Tabs.Panel>
				))}
			</Tabs.Wrapper>
		</PreviewPageContext.Provider>
	);
};

// ------------------------------------------------------------------------------------------------

export const PreviewPageContext = React.createContext<
	| {
			selectedRows: string[];
			setSelectedRows: React.Dispatch<React.SetStateAction<string[]>>;
			filteredSelectedRows: string[];
			runningCodemodsSet: Set<string>;
			setRunningCodemodsSet: React.Dispatch<React.SetStateAction<Set<string>>>;
	  }
	| undefined
>(undefined);
