"use client";

import { Tooltip, useToaster, MenuItem, SplitButton } from "@itwin/itwinui-react";
import { PreviewPageContext, PreviewPageOptimisticData } from "./PreviewPageContent";
import React from "react";
import { SvgCheckmark, SvgScriptRun } from "@itwin/itwinui-icons-react";
import { flushSync } from "react-dom";
import { bulkUpdateAllCodemoddedFiles } from "app/_actions/bulkUpdateCodemoddedFiles";
import { usePersistentToast } from "app/_ui/usePersistentToast";

type SplitButtonType = "bulk-run-and-show-results" | "bulk-run";

export const RunCodemodsInAllFilesButton = ({
	updateDataOptimistically,
}: {
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
}) => {
	const toaster = useToaster();
	const persistentToast = usePersistentToast();

	const previewPageContext = React.useContext(PreviewPageContext);
	const filteredSelectedRows = previewPageContext?.filteredSelectedRows ?? [];

	const isDisabled = (previewPageContext?.runningCodemodsSet.size ?? 0) > 0;

	const action = async () => {
		if (isDisabled) {
			return;
		}

		flushSync(() => {
			previewPageContext?.setRunningCodemodsSet(new Set(filteredSelectedRows));
		});

		const formData = new FormData();
		for (const filePath of previewPageContext?.filteredSelectedRows ?? []) {
			formData.append("file-path", filePath);
		}
		formData.append("new-state", "finished-running");
		formData.append(
			"should-show-results",
			currentButtonType === "bulk-run-and-show-results" ? "true" : "false",
		);

		persistentToast.start(
			`Queued codemods to run on ${previewPageContext?.filteredSelectedRows.length ?? 0} file(s)…`,
		);

		const result = await bulkUpdateAllCodemoddedFiles(formData);

		updateDataOptimistically({ action: "do-nothing", filePath: "" });

		persistentToast.finish();

		if (!result) {
			toaster.positive(
				`Successfully ran codemods for ${previewPageContext?.filteredSelectedRows.length ?? "-"} ${
					previewPageContext?.filteredSelectedRows.length !== 1 ? "files" : "file"
				}`,
			);
		} else {
			toaster.negative("Failed to run codemods in files");
		}

		previewPageContext?.setRunningCodemodsSet(new Set());
	};

	const [currentButtonType, setCurrentButtonType] = React.useState<SplitButtonType>(
		"bulk-run-and-show-results",
	);

	const getButtonMenuItem = (buttonType: SplitButtonType, close: () => void) => {
		const content =
			buttonType === "bulk-run-and-show-results"
				? `Show results after running codemods`
				: "Stay on this page after running codemods";

		const isSelected = currentButtonType === buttonType;

		return (
			<MenuItem
				key={buttonType}
				onClick={() => {
					setCurrentButtonType(buttonType);
					close();
				}}
				isSelected={isSelected}
				endIcon={isSelected ? <SvgCheckmark /> : undefined}
			>
				{content}
			</MenuItem>
		);
	};

	const buttonMenuItems = (close: () => void) => {
		return [
			getButtonMenuItem("bulk-run-and-show-results", close),
			getButtonMenuItem("bulk-run", close),
		];
	};

	const bulkRunButtonContent = `Run codemods in ${filteredSelectedRows.length} ${
		filteredSelectedRows.length !== 1 ? "files" : "file"
	}`;

	const button = (
		<SplitButton
			type="submit"
			styleType="high-visibility"
			disabled={isDisabled}
			startIcon={<SvgScriptRun />}
			menuItems={buttonMenuItems}
		>
			{bulkRunButtonContent}
		</SplitButton>
	);

	const content = isDisabled ? (
		<Tooltip content="Please wait for any queued runs to finish...">{button}</Tooltip>
	) : (
		button
	);

	return !isDisabled ? <form action={action}>{content}</form> : content;
};
