import { Row } from "@itwin/itwinui-react/react-table";
import { SupportedCodemods, supportedCodemods } from "codemods/codemods";
import React from "react";
import { displayToast } from "utils/tasks";
import { CustomDiffEditor } from "./CustomDiffEditor";
import updateCodemoddedFile from "app/_actions/updateCodemoddedFile";
import { PreviewPageContext } from "./PreviewPageContent";
import styles from "./TableExpandableSubComponent.module.scss";
import { CodemodOptions } from "./CodemodOptions";
import { Overlay, ProgressRadial } from "@itwin/itwinui-react";
import type { LatestRunFile } from "utils/latestFileRun";
import { PreviewPageOptimisticData } from "./PreviewPageContent";

export const TableExpandableSubComponent = ({
	optimisticData,
	updateDataOptimistically,
	...row
}: Row<{
	filePath: string;
	name: string;
	relativePath: string;
}> & {
	optimisticData: LatestRunFile;
	updateDataOptimistically: (action: PreviewPageOptimisticData["action"]) => void;
}) => {
	const previewPageContext = React.useContext(PreviewPageContext);
	const file = optimisticData.files.find((f) => f.filePath === row.original.filePath);

	const selectedCodemods =
		file != null
			? file.applicableCodemods.reduce((acc, codemod) => {
					acc[codemod] = file.stages.some((stage) => stage.codemod === codemod);
					return acc;
			  }, {} as Record<typeof supportedCodemods[number], boolean>)
			: ({} as Record<typeof supportedCodemods[number], boolean>);

	if (file == null) {
		return null;
	}

	/**
	 * Runs the codemod after doing the toggle. If the codemod run is successful, then the `selectedCodemods` state is updated.
	 * @param codemodToToggle If "all", then all codemods will be toggled to the value of `value`. Else, only the codemod specified by `codemodToToggle` will be toggled to the value of `value`.
	 */
	const setCodemodCheckboxValue = async (
		codemodToToggle: SupportedCodemods | "all",
		value: boolean,
	) => {
		if (previewPageContext?.runningCodemodsSet.has(file.filePath)) {
			return;
		}

		updateDataOptimistically({
			action: "do-nothing",
			filePath: "",
		});

		previewPageContext?.setRunningCodemodsSet((prev) => {
			const newSet = new Set(prev);
			newSet.add(file.filePath);
			return newSet;
		});

		let newSelectedCodemods: Record<SupportedCodemods, boolean>;
		if (codemodToToggle === "all") {
			newSelectedCodemods = supportedCodemods.reduce((acc, codemod) => {
				acc[codemod] = value ?? !selectedCodemods[codemod];
				return acc;
			}, {} as Record<SupportedCodemods, boolean>);
		} else {
			newSelectedCodemods = {
				...selectedCodemods,
				[codemodToToggle]: value ?? !selectedCodemods[codemodToToggle],
			};
		}

		const codemodsToRun = Object.entries(newSelectedCodemods);

		const formData = new FormData();
		formData.append("file-path", file.filePath);
		for (const [codemod, selected] of codemodsToRun) {
			if (selected) {
				formData.append("codemods-to-run", codemod);
			}
		}

		const newLatestRunFile = await updateCodemoddedFile(formData);

		if (newLatestRunFile == null) {
			displayToast("negative", "Failed to re-run codemods");
		}

		previewPageContext?.setRunningCodemodsSet((prev) => {
			const newSet = new Set(prev);
			newSet.delete(file.filePath);
			return newSet;
		});
	};

	const content = (
		<div className={styles.expandableContent}>
			<CodemodOptions
				file={file}
				selectedCodemods={selectedCodemods}
				setCodemodCheckboxValue={setCodemodCheckboxValue}
				optimisticData={optimisticData}
				updateDataOptimistically={updateDataOptimistically}
			/>
			<CustomDiffEditor
				originalCode={file.originalFileContents}
				modifiedCode={
					file.stages.length > 0
						? file.stages[file.stages.length - 1].output
						: file.originalFileContents
				}
			/>
		</div>
	);

	return previewPageContext?.runningCodemodsSet.has(file.filePath) ? (
		<Overlay.Wrapper className={styles.overlay}>
			<Overlay.Overlay>
				<ProgressRadial indeterminate />
			</Overlay.Overlay>
			<Overlay.HiddenContent>{content}</Overlay.HiddenContent>
		</Overlay.Wrapper>
	) : (
		content
	);
};
