import { Flex, ProgressRadial, useToaster } from "@itwin/itwinui-react";
import React from "react";

const wait = (ms: number) => new Promise((resolve) => setTimeout(resolve, ms));

// Debounce helpers
const [runAfterDelay, cancel] = (() => {
	let running = false;

	return [
		async (run: Function) => {
			running = true;
			await wait(300);
			if (!running) return;
			run();
		},
		() => {
			running = false;
		},
	];
})();

// ------------------------------------------------------------------------------------------------

/**
 * Debounced persistent toast. Call `start()` to show the toast, and `finish()` to hide it.
 *
 * If `finish()` is called before the debounce delay, then the toast will not be shown.
 */
export function usePersistentToast() {
	const toaster = useToaster();
	const closeRunningToast = React.useRef(() => {});

	return React.useMemo(
		() => ({
			start: (message: string) => {
				runAfterDelay(() => {
					const { close } = toaster.informational(
						<Flex>
							<ProgressRadial size="x-small" />
							{message}
						</Flex>,
						{ type: "persisting" },
					);
					closeRunningToast.current = close;
				});
			},
			finish: () => {
				cancel();
				closeRunningToast.current();
				closeRunningToast.current = () => {};
			},
		}),
		[toaster],
	);
}
