import type { API, FileInfo, Options } from "jscodeshift";
import { getSafeCodemoddedSource, getImportAliasToRealComponentMapping } from "../utils/codemods";

/**
 * Codemod for the below change:
 * > #### Breadcrumbs
 * > - [#1557](https://github.com/iTwin/iTwinUI/pull/1557): Directly passing `<a>`/`<Button>`/`<span>` as children is deprecated. Should be replaced with `Breadcrumbs.Item` subcomponent.
 */
const breadcrumbsChildren = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	// Find all Breadcrumbs
	const allBreadcrumbsImports = getImportAliasToRealComponentMapping(j, root, ["Breadcrumbs"]);

	Object.entries(allBreadcrumbsImports).forEach(([alias, realName]) => {
		const allBreadcrumbs = root.findJSXElements(alias);

		// Find all `a`/`Button`/`span` descendants of Breadcrumbs
		// The descendants could be nested in the AST. E.g. using a loop to create the items from a list.
		const allChildrenToRename = [
			...allBreadcrumbs.findJSXElements("a").nodes(),
			...allBreadcrumbs.findJSXElements("Button").nodes(),
			...allBreadcrumbs.findJSXElements("span").nodes(),
		];

		// Rename each JSXElement in allChildrenToRename to Breadcrumbs.Item
		allChildrenToRename.forEach((child) => {
			if (child.openingElement.name.type === "JSXIdentifier") {
				child.openingElement.name.name = `${alias}.Item`;
				hasChanged = true;
			}
			if (child.closingElement?.name.type === "JSXIdentifier") {
				child.closingElement.name.name = `${alias}.Item`;
				hasChanged = true;
			}
		});
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default breadcrumbsChildren;
