import type { API, FileInfo, Options } from "jscodeshift";
import {
	getPropFromComponent,
	getSafeCodemoddedSource,
	moveClassNameAndStylePropsToWrapperProps,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1371](https://github.com/iTwin/iTwinUI/pull/1371): `className` and `style` will now always be applied on the checkbox input element (instead of conditionally being applied on the wrapper in some cases). Added new `wrapperProps` and `labelProps` to allow for styling of wrapping label and label text.
 * > ```diff
 * > <Checkbox
 * >     label = 'Checkbox label'
 * > -   className={wrapperClassname}
 * > -   style={wrapperStyle}
 * > +   wrapperProps={{ className: wrapperClassName, style: wrapperStyle }}
 * > />
 * > ```
 */
const checkboxMoveStyleClassInside = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	const hasChanged = moveClassNameAndStylePropsToWrapperProps({
		j,
		root,
		componentName: "Checkbox",
		// excludeCondition: If label is not passed, then we don't need to do anything, because there is no wrapper
		excludeCondition: (path) => getPropFromComponent(path, "label") == null,
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default checkboxMoveStyleClassInside;
