import { API, FileInfo, Options } from "jscodeshift";
import {
	addSuggestionToSuggestionsFile,
	getImportAliasToRealComponentMapping,
	getITwinUIReactImportDeclarations,
	getSafeCodemoddedSource,
} from "../utils/codemods";

const deprecatedComponentsToReplace = [
	{
		name: "Body",
		as: "p",
	},
	{
		name: "Headline",
		as: "h1",
	},
	{
		name: "Leading",
		as: "h3",
	},
	{
		name: "Small",
		as: "small",
	},
	{
		name: "Subheading",
		as: "h3",
	},
	{
		name: "Title",
		as: "h2",
	},
] as const;

/**
 * Codemod for the below change:
 * > [#1296](https://github.com/iTwin/iTwinUI/pull/1296): Removed previously-deprecated typography components (`Body`, `Headline`, `Leading`, `Small`, `Subheading`, `Title`) which have been replaced by `Text`.
 */
const deprecatedTypography = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	try {
		const allITwinUIDeprecatedTypographyComponents = getImportAliasToRealComponentMapping(
			j,
			root,
			deprecatedComponentsToReplace.map((dc) => dc.name),
		);

		//----------------------------------------------
		// Replace all aliases of deprecated components with real names

		Object.keys(allITwinUIDeprecatedTypographyComponents).forEach((alias) => {
			const realName = allITwinUIDeprecatedTypographyComponents[alias];

			// Imports
			getITwinUIReactImportDeclarations(j, root).forEach((importDeclaration) => {
				importDeclaration.node.specifiers?.forEach((specifier) => {
					if (specifier.local?.name === alias) {
						specifier.local.name = realName;
						hasChanged = true;
					}
				});
			});

			// JSX Elements
			root.findJSXElements(alias).forEach(({ node: component }) => {
				if (component.openingElement.name.type == "JSXIdentifier") {
					component.openingElement.name.name = realName;
					hasChanged = true;
				}
				if (component.closingElement?.name.type == "JSXIdentifier") {
					component.closingElement.name.name = realName;
					hasChanged = true;
				}
			});
		});

		//----------------------------------------------
		// Replace all deprecated components with Text

		let isAtleastOneReplacementMade = false;

		Object.entries(allITwinUIDeprecatedTypographyComponents).forEach(
			([componentToReplaceAlias, componentToReplaceRealName]) => {
				const componentToReplace = deprecatedComponentsToReplace.find(
					(c) => c.name === componentToReplaceRealName,
				);
				if (componentToReplace == null) {
					return;
				}

				const foundComponents = root.findJSXElements(componentToReplace.name);
				if (foundComponents.length == 0) {
					return;
				}

				isAtleastOneReplacementMade = true;

				foundComponents.forEach(({ node: component }) => {
					// Replace component with Text
					if (component.openingElement.name.type == "JSXIdentifier") {
						component.openingElement.name.name = "Text";
						hasChanged = true;
					}
					if (component.closingElement?.name.type == "JSXIdentifier") {
						component.closingElement.name.name = "Text";
						hasChanged = true;
					}

					// Add variant="component" to Text to the beginning of the attributes
					component.openingElement.attributes?.unshift(
						j.jsxAttribute(
							j.jsxIdentifier("variant"),
							j.literal(componentToReplace.name.toLowerCase()),
						),
					);
					hasChanged = true;

					// If !isTagAHeading, add as="..." to Text
					if (!isTagAHeading(componentToReplace.as)) {
						component.openingElement.attributes?.unshift(
							j.jsxAttribute(j.jsxIdentifier("as"), j.literal(componentToReplace.as)),
						);
						hasChanged = true;
					}
					// Add a one-time suggestion to confirm the heading levels (i.e. whether as="..." is needed)
					else {
						const componentName = componentToReplace.name;
						const componentVariant = componentToReplace.name.toLowerCase();

						const message = `The \`<${componentName}>...</${componentName}>\` has been automatically replaced with \`<Text variant="${componentVariant}">...</Text>\`. However, since \`<${componentName}>\` was a \`<${componentToReplace.as}>\` and \`<Text>\` is always a \`<div>\`, please reconfirm your heading levels. You may need to manually pass an \`as="..."\` to \`<Text>\`. E.g. \`<Text as=\"${componentToReplace.as}\" variant="${componentVariant}">...</Text>\`. ([MDN Docs](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/Heading_Elements#accessibility_concerns))`;
						addSuggestionToSuggestionsFile({
							fileInfo,
							loc: component.loc,
							message,
							codemod: "deprecated-typography",
						});
					}
				});
			},
		);

		//----------------------------------------------
		// Update imports

		// If no replacement has been made, don't make any changes to the imports
		// Throw anything to skip subsequent steps
		if (!isAtleastOneReplacementMade) {
			throw null;
		}

		// Find all imports from @itwin/itwinui-react
		let importDeclarations = getITwinUIReactImportDeclarations(j, root);

		if (importDeclarations.length === 0) {
			// If no imports from @itwin/itwinui-react, add one
			importDeclarations
				.at(0)
				.insertBefore(j.importDeclaration([], j.literal("@itwin/itwinui-react")));
			hasChanged = true;
		}

		// Check if Text is imported in one of the ImportDeclarations
		let textImportAlreadyExists = false;
		importDeclarations.forEach((importDeclaration) => {
			const textImportSpecifier = importDeclaration.node.specifiers?.find(
				(specifier) => specifier.local?.name === "Text",
			);
			if (textImportSpecifier != null) {
				textImportAlreadyExists = true;
			}
		});

		// If !textImportAlreadyExists, add import of Text to the first ImportDeclaration in importDeclarations
		if (!textImportAlreadyExists) {
			importDeclarations.nodes()[0].specifiers?.push(j.importSpecifier(j.identifier("Text")));
			hasChanged = true;
		}

		importDeclarations.forEach((importDeclaration) => {
			if (importDeclaration.node.specifiers == null) {
				return;
			}

			// Remove import of deprecated components and their aliases
			const newImportSpecifiers = importDeclaration.node.specifiers.filter((specifier) => {
				const deprecatedComponentsToReplaceStringArray = deprecatedComponentsToReplace.map(
					(c) => c.name,
				) as unknown as string[];

				// E.g. #1: import { Body as BodyOld } from "@itwin/itwinui-react";
				// E.g. #2: import { Body } from "@itwin/itwinui-react";
				const specifierName = specifier.local?.name; // #1: BodyOld | #2: Body
				const specifierImportedName = (specifier as any)?.imported?.name; // #1: Body | #2: Body

				const isSpecifierNameDeprecated = deprecatedComponentsToReplaceStringArray.includes(
					specifierName ?? "",
				);
				const isSpecifierImportedNameDeprecated = deprecatedComponentsToReplaceStringArray.includes(
					specifierImportedName ?? "",
				);

				return !isSpecifierNameDeprecated && !isSpecifierImportedNameDeprecated;
			});

			if (importDeclaration.node.specifiers.length !== newImportSpecifiers.length) {
				hasChanged = true;
			}

			importDeclaration.node.specifiers = newImportSpecifiers;
		});

		//----------------------------------------------
		// Remove empty import declarations

		// Find all imports from @itwin/itwinui-react
		importDeclarations = getITwinUIReactImportDeclarations(j, root);

		importDeclarations.forEach((importDeclaration) => {
			if (importDeclaration.node.specifiers?.length === 0) {
				importDeclaration.prune();
				hasChanged = true;
			}
		});
	} finally {
		return getSafeCodemoddedSource({
			root,
			fileInfo,
			hasChanged,
		});
	}
};

const isTagAHeading = (tag: keyof React.JSX.IntrinsicElements) => {
	return ["h1", "h2", "h3", "h4", "h5", "h6"].includes(tag);
};

export default deprecatedTypography;
