import type {
	API,
	FileInfo,
	Options,
	ASTPath,
	ImportDeclaration,
	ImportSpecifier,
} from "jscodeshift";
import {
	getITwinUIReactImportDeclarations,
	getImportAliasToRealComponentMapping,
	getImportSpecifierName,
	getSafeCodemoddedSource,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1298](https://github.com/iTwin/iTwinUI/pull/1298): `Props` types will no longer be exported for any component.
 * > Use `React.ComponentProps` or `React.ComponentPropsWithoutRef` instead.
 * >   ```diff
 * >   - import { Avatar, type AvatarProps } from '@itwin/itwinui-react';
 * >   + import { Avatar } from '@itwin/itwinui-react';
 * >   + type AvatarProps = React.ComponentPropsWithoutRef<typeof Avatar>;
 * >   ```
 * @param fileInfo
 * @param api
 * @param options
 * @returns
 */
const removeImportProps = (fileInfo: FileInfo, api: API, options: Options): string => {
	// Logic:
	// - For each `ImportDeclaration`, we go through each `ImportSpecifier`s
	// - If the ImportSpecifier is a props import (i.e. ends with `Props`),
	//   we remove it from the `ImportDeclaration` and add it to the `allTypeImports` array
	// - Then for each import in `allTypeImports`, we add something of the form:
	//   `type MyComponentProps = React.ComponentPropsWithoutRef<typeof MyComponent>;`
	// - If the `ImportDeclaration` is empty after removing the props imports,
	//   we add it to the `importDeclarationsToRemove` array
	// - Then we remove all the `ImportDeclaration`s in `importDeclarationsToRemove`

	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	const importDeclarations = getITwinUIReactImportDeclarations(j, root);
	const importDeclarationsToRemove: ASTPath<ImportDeclaration>[] = [];
	const allImportedComponentsProps: Array<ReturnType<typeof getImportSpecifierName>> = [];

	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root);

	importDeclarations.forEach((path) => {
		const node = path.node as ImportDeclaration;

		const specifiers = node.specifiers!.filter(
			(specifier) =>
				specifier.type === "ImportSpecifier" && specifier.imported.name.endsWith("Props"),
		);

		// If there are any props imports
		if (specifiers.length > 0) {
			// All the ImportSpecifiers to not remove from ImportDeclarations
			const allNonPropImportSpecifiers: ImportSpecifier[] = [];

			// Go through all import specifiers (not just props import specifiers)
			node.specifiers!.forEach((specifier) => {
				if (specifier.type === "ImportSpecifier") {
					const specifierNames = getImportSpecifierName(specifier);

					if (!specifierNames.importedName?.endsWith("Props")) {
						// E.g. Avatar, Button, ...
						const nonPropsImportedSpecifier = specifier;
						allNonPropImportSpecifiers.push(nonPropsImportedSpecifier);
					} else {
						// E.g. AvatarProps, ButtonProps, AvatarPropsAlias, ButtonPropsAlias ...
						allImportedComponentsProps.push(specifierNames);
					}
				}
			});

			// if allNonPropImportSpecifiers is empty
			// we want to remove the import declaration
			if (allNonPropImportSpecifiers.length === 0) {
				importDeclarationsToRemove.push(path);
			}
			// Else, we want to update the specifiers to only include all the non-props imports
			else {
				node.specifiers = allNonPropImportSpecifiers;
				hasChanged = true;
			}
		}
	});

	// Remove all the empty import declarations
	importDeclarationsToRemove.forEach((path) => {
		path.prune();
		hasChanged = true;
	});

	// Reverse the allTypeImports array since the below code adds them in reverse order
	allImportedComponentsProps.reverse();

	const imports = getITwinUIReactImportDeclarations(j, root);
	// For each allPropImportSpecifiers, declare it as a type at the bottom of all imports
	// or the beginning of the file if there are no imports
	allImportedComponentsProps.forEach((componentProps) => {
		const componentName = componentProps.importedName?.replace("Props", ""); // E.g. Avatar

		// Find if componentName is already imported
		let importedComponentName: string | undefined;
		Object.entries(importAliasToRealComponentMapping).forEach(([alias, realName]) => {
			if (realName === componentName) {
				importedComponentName = alias;
			}
		});

		// If componentName is not imported, add an import specifier for it
		if (importedComponentName == null && componentName != null) {
			const newImportSpecifier = j.importSpecifier(j.identifier(componentName));
			imports
				.at(imports.length - 1)
				.get()
				.node.specifiers.push(newImportSpecifier);
			hasChanged = true;

			importedComponentName = componentName;
		}

		const componentType = `React.ComponentPropsWithoutRef<typeof ${importedComponentName}>`;
		const componentTypeDeclaration = `type ${componentProps.name} = ${componentType};`;

		if (imports.length > 0) {
			// Add after all the imports
			j(imports.at(imports.length - 1).get()).insertAfter(componentTypeDeclaration);
			hasChanged = true;
		} else {
			// Add at the beginning of the file if no imports
			root.get().node.program.body.unshift(componentTypeDeclaration);
			hasChanged = true;
		}
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default removeImportProps;
