import { API, ASTPath, FileInfo, JSXOpeningElement, Options } from "jscodeshift";
import {
	addSuggestionToSuggestionsFile,
	filterOutObjectPropertyFromObjectExpression,
	filterOutPropFromComponent,
	getImportAliasToRealComponentMapping,
	getObjectPropertyFromObjectExpression,
	getPropFromComponent,
	getSafeCodemoddedSource,
} from "../utils/codemods";

/**
 * Codemod for the below change (slightly modified from the original migration guide version):
 * > [#1406](https://github.com/iTwin/iTwinUI/pull/1406): Removed `setFocus` prop. The component can be focused by using `ref`.
 * > Removed `setFocus` from:
 * > - Checkbox
 * > - ColorPicker
 * > - ComboBox
 * > - Input
 * > - LabeledInput
 * > - LabeledTextarea
 * > - Radio
 * > - RadioTile
 * > - Select
 * > - Slider
 * > - Textarea
 * > - ToggleSwitch
 * >
 * > Added setFocus functionality to:
 * > - TextFilter
 * > - NumberRangeFilter
 * > - DateRangeFilter
 * >
 * > Keeping setFocus on (because of more complex use case):
 * > - DatePicker
 * > - Dialog
 * > - Menu
 * > - TimePicker
 * @param fileInfo
 * @param api
 * @param options
 */
const setFocusRemoved = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	const componentsWithSetFocusRemoved = [
		"Checkbox",
		"ColorPicker",
		"ComboBox",
		"Input",
		"LabeledInput",
		"LabeledTextarea",
		"Radio",
		"RadioTile",
		"Select",
		"Slider",
		"Textarea",
		"ToggleSwitch",
	];

	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root);

	for (const componentWithSetFocusRemoved of componentsWithSetFocusRemoved) {
		for (const [componentName, realComponentName] of Object.entries(
			importAliasToRealComponentMapping,
		)) {
			if (realComponentName === componentWithSetFocusRemoved) {
				root
					.find(j.JSXOpeningElement, {
						name: {
							type: "JSXIdentifier",
							name: componentName,
						},
					})
					.forEach((path) => {
						const newHasChanged = removeSetFocusProp({
							j,
							fileInfo,
							path,
						});
						hasChanged = hasChanged || newHasChanged;
					});
			}
		}
	}

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

// ------------------------------------------------------------------------------------------------

/**
 * Remove usage of setFocus prop from components that don't support it.
 * @returns hasChanged
 */
const removeSetFocusProp = ({
	j,
	fileInfo,
	path,
}: {
	j: API["jscodeshift"];
	fileInfo: FileInfo;
	path: ASTPath<JSXOpeningElement>;
}): boolean => {
	let hasChanged = false;

	try {
		const componentName = path.node.name.type === "JSXIdentifier" ? path.node.name.name : null;
		if (componentName == null) {
			throw null;
		}

		// Cases:
		// 1. If component == ComboBox, setFocus is passed in the inputProps as `inputProps: {setFocus: true}`
		// 2. Else, setFocus is passed as a prop as `setFocus={true}`

		// Case 1: ComboBox
		if (componentName === "ComboBox") {
			const newHasChanged = removeSetFocusPropFromComboBox({
				fileInfo,
				path,
			});
			hasChanged = hasChanged || newHasChanged;
		}

		// Case 2: All other components
		else {
			const newHasChanged = removeSetFocusPropFromNonComboBoxComponents({
				j,
				fileInfo,
				path,
			});
			hasChanged = hasChanged || newHasChanged;
		}
	} finally {
		return hasChanged;
	}
};

const removeSetFocusPropFromComboBox = ({
	fileInfo,
	path,
}: {
	fileInfo: FileInfo;
	path: ASTPath<JSXOpeningElement>;
}): boolean => {
	let hasChanged = false;

	try {
		if (path.node.name.type !== "JSXIdentifier" || path.node.name.name !== "ComboBox") {
			throw null;
		}

		const inputProps = getPropFromComponent(path, "inputProps");

		if (
			inputProps != null &&
			inputProps.value != null &&
			inputProps.value &&
			inputProps.value.type === "JSXExpressionContainer" &&
			inputProps.value.expression != null &&
			inputProps.value.expression.type === "ObjectExpression"
		) {
			const inputPropsObjectExpression = inputProps.value.expression;
			const setFocusAttribute = getObjectPropertyFromObjectExpression(
				inputPropsObjectExpression,
				"setFocus",
			);

			// If setFocus={false}, it can be auto-removed
			if (
				setFocusAttribute?.value.type === "BooleanLiteral" &&
				setFocusAttribute.value.value === false
			) {
				const newInputPropsObjectExpression = filterOutObjectPropertyFromObjectExpression(
					inputPropsObjectExpression,
					"setFocus",
				);
				if (newInputPropsObjectExpression.length !== inputPropsObjectExpression.properties.length) {
					inputProps.value.expression.properties = newInputPropsObjectExpression;
					hasChanged = true;
				}
			}

			// If setFocus={true}, replace `setFocus` with `autoFocus`
			else if (
				setFocusAttribute?.value.type === "BooleanLiteral" &&
				setFocusAttribute.value.value === true
			) {
				if (
					setFocusAttribute.key.type === "Identifier" &&
					setFocusAttribute.key.name === "setFocus"
				) {
					// Rename setFocus to autoFocus
					setFocusAttribute.key.name = "autoFocus";
				}
			}

			// If is not of the form setFocus={boolean}, no automated fix is possible
			else {
				// Give the user a manual warning that they need to remove the setFocus prop and use a ref instead.
				addSuggestionToSuggestionsFile({
					codemod: "setfocus-removed",
					fileInfo,
					loc: setFocusAttribute?.loc ?? inputProps?.loc,
					message: suggestionMessage("ComboBox"),
				});
			}
		}
	} finally {
		return hasChanged;
	}
};

const removeSetFocusPropFromNonComboBoxComponents = ({
	j,
	fileInfo,
	path,
}: {
	j: API["jscodeshift"];
	fileInfo: FileInfo;
	path: ASTPath<JSXOpeningElement>;
}): boolean => {
	let hasChanged = false;

	try {
		if (path.node.attributes == null) {
			throw null;
		}

		const componentName = path.node.name.type === "JSXIdentifier" ? path.node.name.name : null;
		if (componentName == null) {
			throw null;
		}

		const setFocusAttribute = getPropFromComponent(path, "setFocus");
		if (setFocusAttribute != null) {
			// If setFocus={boolean} (i.e. setFocus={true/false} and not setFocus={someVariable})
			if (
				setFocusAttribute.value?.type === "JSXExpressionContainer" &&
				setFocusAttribute.value.expression.type === "BooleanLiteral"
			) {
				// if setFocus={false}, it can be auto-removed
				if (setFocusAttribute.value.expression.value === false) {
					const newAttributes = filterOutPropFromComponent(path, "setFocus");

					if (newAttributes.length !== path.node.attributes.length) {
						hasChanged = true;
						path.node.attributes = newAttributes;
					}
				}

				// If setFocus={true},
				// 1. If component == ColorPicker, Slider, just show a manual suggestion since these are complex cases
				// 2. If component == Select, replace `setFocus` with `triggerProps={autoFocus: true}`
				// 3. Else, replace `setFocus` with `autoFocus`
				else {
					// Case 1
					if (componentName === "ColorPicker" || componentName === "Slider") {
						addSuggestionToSuggestionsFile({
							codemod: "setfocus-removed",
							fileInfo,
							loc: setFocusAttribute.loc,
							message: suggestionMessage(componentName),
						});
					}

					// Case 2
					else if (componentName === "Select") {
						// Remove setFocus prop
						path.node.attributes = filterOutPropFromComponent(path, "setFocus");

						// Add `triggerProps={ref: React.useCallback((el) => el?.focus(), [])}`
						(() => {
							let triggerProps = getPropFromComponent(path, "triggerProps");

							// Create empty triggerProps if it doesn't exist
							if (triggerProps == null) {
								path.node.attributes.push(
									j.jsxAttribute(
										j.jsxIdentifier("triggerProps"),
										j.jsxExpressionContainer(j.objectExpression([])),
									),
								);

								hasChanged = true;

								triggerProps = getPropFromComponent(path, "triggerProps");
							}

							const triggerPropsObjectExpression =
								triggerProps?.value?.type === "JSXExpressionContainer" &&
								triggerProps.value.expression.type === "ObjectExpression"
									? triggerProps.value.expression
									: null;

							if (triggerPropsObjectExpression != null) {
								// If autoFocus does not exist in triggerProps, add it
								const ref = getObjectPropertyFromObjectExpression(
									triggerPropsObjectExpression,
									"ref",
								);

								if (ref == null) {
									const newRefObjectProperty = j.objectProperty(
										j.identifier("ref"),
										j.callExpression(
											j.memberExpression(j.identifier("React"), j.identifier("useCallback")),
											[
												j.arrowFunctionExpression(
													[j.identifier("el")],
													j.callExpression(
														j.optionalMemberExpression(j.identifier("el"), j.identifier("focus")),
														[],
													),
												),
												j.arrayExpression([]),
											],
										),
									);
									triggerPropsObjectExpression.properties.push(newRefObjectProperty);
									hasChanged = true;
								} else {
									addSuggestionToSuggestionsFile({
										codemod: "setfocus-removed",
										fileInfo,
										loc: setFocusAttribute.loc,
										message: suggestionMessage(componentName),
									});
								}
							}
						})();
					}

					// Case 3
					else {
						// Rename setFocus to autoFocus
						setFocusAttribute.name.name = "autoFocus";
					}
				}
			}

			// If is not of the form setFocus={boolean}, no automated fix is possible
			else {
				// Give the user a manual warning that they need to remove the setFocus prop and use a ref instead.
				addSuggestionToSuggestionsFile({
					codemod: "setfocus-removed",
					fileInfo,
					loc: setFocusAttribute.loc,
					message: suggestionMessage(componentName),
				});
			}
		}
	} finally {
		return hasChanged;
	}
};

const suggestionMessage = (componentName: string) => {
	const comboBoxComponentMessage = `The \`inputPros.setFocus\` prop has been removed from \`ComboBox\`. Please delete the \`inputProps.setFocus\` prop and instead use a \`inputProps.ref\` or \`inputProps.autoFocus\` to set the focus when/if needed.

Examples:

* Implicitly focus when the component mounts:
	\`\`\`tsx
	<ComboBox inputProps={{ autoFocus: true }} />
	\`\`\`

* Explicitly focus when the component mounts:
	\`\`\`tsx
	<ComboBox
		inputProps={{
			// don't forget the useCallback!
			ref: React.useCallback((el) => el?.focus(), [])
		}}
	/>
	\`\`\`
`.trim();

	const selectComponentMessage = `The \`triggerProps.setFocus\` prop has been removed from \`Select\`. Please delete the \`triggerProps.setFocus\` prop and instead use a \`triggerProps.ref\` or \`triggerProps.autoFocus\` to set the focus when/if needed.

Examples:

* Implicitly focus when the component mounts:
	\`\`\`tsx
	<Select triggerProps={{ autoFocus: true }} />
	\`\`\`

* Explicitly focus when the component mounts:
	\`\`\`tsx
	<Select
		triggerProps={{
			// don't forget the useCallback!
			ref: React.useCallback((el) => el?.focus(), [])
		}}
	/>
	\`\`\`
`.trim();

	const sliderComponentMessage = `The \`setFocus\` prop has been removed from \`Slider\`. Please delete the \`setFocus\` prop and instead use a \`ref\` or \`autoFocus\` to set the focus on the thumb when/if needed.

Examples:

* Implicitly focus when the component mounts:
	\`\`\`tsx
	<Slider thumbProps={() => ({ autoFocus: true })} />
	\`\`\`

* Explicitly focus when the component mounts:
	\`\`\`tsx
	const autoFocusRef = React.useCallback(el => el?.focus(), []);
	<Slider thumbProps={() => ({ ref: autoFocusRef })} />
	\`\`\`

* When the Slider has multiple thumbs:
	\`\`\`tsx
	const autoFocusRef = React.useCallback(el => el?.focus(), []);
	<Slider thumbProps={(index) => index === 0 ? ({ ref: autoFocusRef }) : ({})} />
	\`\`\`
`.trim();

	const colorPickerComponentMessage = `The \`setFocus\` prop has been removed from \`ColorPicker\`. Please delete the \`setFocus\` prop and instead wrap the \`ColorPicker\` with a \`Popover\` to control focus.
	
\`\`\`tsx
<Popover content={<ColorPicker />} />
\`\`\`
`.trim();

	const easyCasesComponentMessage = `The \`setFocus\` prop has been removed from \`${componentName}\`. Please delete the \`setFocus\` prop and instead use a ref or \`autoFocus\` to set the focus when/if needed.

Examples:

* Implicitly focus when the component mounts:
	\`\`\`tsx
	<${componentName} autoFocus />
	\`\`\`

* Explicitly focus when the component mounts:
	\`\`\`tsx
	<${componentName}
		// don't forget the useCallback!
		ref={React.useCallback((el) => el?.focus(), [])}
	/>
  \`\`\`
`.trim();

	if (componentName === "ComboBox") {
		return comboBoxComponentMessage;
	}
	if (componentName === "Select") {
		return selectComponentMessage;
	}
	if (componentName === "ColorPicker") {
		return colorPickerComponentMessage;
	}
	if (componentName === "Slider") {
		return sliderComponentMessage;
	}
	return easyCasesComponentMessage;
};

export default setFocusRemoved;
