import { API, FileInfo, JSXAttribute, JSXExpressionContainer, Options } from "jscodeshift";
import {
	getImportAliasToRealComponentMapping,
	getPropFromComponent,
	getSafeCodemoddedSource,
	getSimplifiedJSXAttributeValue,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1435](https://github.com/iTwin/iTwinUI/pull/1435): SplitButton: `className` is now passed to the button instead of the component wrapper. Use `wrapperProps` if the old behavior is desired.
 */
const splitButtonClassName = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root, [
		"SplitButton",
	]);
	Object.keys(importAliasToRealComponentMapping).forEach((splitButtonAlias) => {
		const splitButtons = root.findJSXElements(splitButtonAlias);
		splitButtons.forEach((splitButton) => {
			if (splitButton.node.openingElement.attributes == null) {
				return;
			}

			const className = splitButton.node.openingElement.attributes.find(
				(attr) => attr.type === "JSXAttribute" && attr.name.name === "className",
			) as JSXAttribute | undefined;

			if (className != null) {
				const wrapperProps = getPropFromComponent(splitButton, "wrapperProps")
					?.value as JSXExpressionContainer;

				if (wrapperProps != null) {
					// We should do nothing, since looks like the user already knows about wrapperProps.
					// i.e. The user really wants to pass the className to the Button and not the wrapper.
				} else {
					const classNameValue = getSimplifiedJSXAttributeValue(className);
					if (classNameValue != null) {
						// Add wrapperProps (with className) to SplitButton
						splitButton.node.openingElement.attributes.push(
							j.jsxAttribute(
								j.jsxIdentifier("wrapperProps"),
								j.jsxExpressionContainer(
									j.objectExpression([j.objectProperty(j.identifier("className"), classNameValue)]),
								),
							),
						);
						hasChanged = true;

						// Remove className JSXAttribute from SplitButton
						const newAttributes = splitButton.node.openingElement.attributes.filter(
							(attr) => attr !== className,
						);

						if (newAttributes.length !== splitButton.node.openingElement.attributes.length) {
							splitButton.node.openingElement.attributes = newAttributes;
							hasChanged = true;
						}
					}
				}
			}
		});
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default splitButtonClassName;
