import { API, FileInfo, JSCodeshift, JSXAttribute, Options } from "jscodeshift";
import {
	addSuggestionToSuggestionsFile,
	getImportAliasToRealComponentMapping,
	getSafeCodemoddedSource,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1527](https://github.com/iTwin/iTwinUI/pull/1527): Tile's `leftIcon` and `rightIcon` props no longer set `size` for the buttons so it must be done manually.
 * > ```diff
 * >   <Tile
 * >     // ...
 * > -  leftIcon={<IconButton styleType='borderless'>...</IconButton>}
 * > +  leftIcon={<IconButton styleType='borderless' size='small'>...</IconButton>}
 * > -  rightIcon={<IconButton styleType='borderless'>...</IconButton>}
 * > +  rightIcon={<IconButton styleType='borderless' size='small'>...</IconButton>}
 * >   />
 * > ```
 * @param fileInfo
 * @param api
 * @param options
 */
const tileIconButtonSize = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	const tileAliasComponents = getImportAliasToRealComponentMapping(j, root);
	Object.keys(tileAliasComponents).forEach((tileAliasComponent) => {
		const tileComponents = root.findJSXElements(tileAliasComponent);
		tileComponents.forEach((element) => {
			const tileComponent = element.node;

			// Find leftIcon & rightIcon that are `JSXAttribute`s
			// We are not handling `JSXSpreadAttribute`s because handling non JSXAttributes would be too complex for a codemod.
			// For non `JSXAttribute`s, we report this as a required manual change.
			const leftIcon = tileComponent.openingElement.attributes?.find(
				(attr) => attr.type == "JSXAttribute" && attr.name.name === "leftIcon",
			) as JSXAttribute | undefined;
			const rightIcon = tileComponent.openingElement.attributes?.find(
				(attr) => attr.type == "JSXAttribute" && attr.name.name === "rightIcon",
			) as JSXAttribute | undefined;

			if (leftIcon) {
				const newHasChanged = addDefaultSizeAttributeToIconButtonPassedAsValue(
					leftIcon,
					j,
					root,
					fileInfo,
				);
				hasChanged = hasChanged || newHasChanged;
			}
			if (rightIcon) {
				const newHasChanged = addDefaultSizeAttributeToIconButtonPassedAsValue(
					rightIcon,
					j,
					root,
					fileInfo,
				);
				hasChanged = hasChanged || newHasChanged;
			}
		});
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

/**
 * Adds the size="small" attribute to the IconButton if the size attribute doesn't exist. This is used for both leftIcon and rightIcon.
 * This works only when the JSXAttribute's value is an `<IconButton>`.
 */
const addDefaultSizeAttributeToIconButtonPassedAsValue = (
	jsxAttribute: JSXAttribute,
	j: JSCodeshift,
	root: ReturnType<typeof j>,
	fileInfo: FileInfo,
): boolean => {
	let hasChanged = false;

	const iconButtonAliasComponents = getImportAliasToRealComponentMapping(j, root, ["IconButton"]);
	Object.keys(iconButtonAliasComponents).forEach((iconButtonAliasComponent) => {
		const iconAttributeValue = jsxAttribute.value;

		// If iconAttributeValue is an IconButton, then add the size="small" attribute if the size attribute doesn't exist
		if (
			iconAttributeValue != null &&
			iconAttributeValue.type === "JSXExpressionContainer" &&
			iconAttributeValue.expression.type === "JSXElement" &&
			iconAttributeValue.expression.openingElement.name.type == "JSXIdentifier" &&
			iconAttributeValue.expression.openingElement.name.name === iconButtonAliasComponent
		) {
			const iconButton = iconAttributeValue.expression;

			const size = iconButton.openingElement.attributes?.find(
				(attr) => attr.type == "JSXAttribute" && attr.name.name === "size",
			);
			if (!size) {
				iconButton.openingElement.attributes?.push(
					j.jsxAttribute(j.jsxIdentifier("size"), j.literal("small")),
				);
				hasChanged = true;
			}
		} else {
			addSuggestionToSuggestionsFile({
				codemod: "tile-iconbutton-size",
				fileInfo,
				loc: jsxAttribute.loc,
				message: `\`leftIcon\`/\`rightIcon\` was not passed an \`<IconButton>\` directly. Manually verify whether \`size="small"\` needs to be added to the indirect \`<IconButton>\` passed to \`leftIcon\`/\`rightIcon\``,
			});
		}
	});

	return hasChanged;
};

export default tileIconButtonSize;
