import type { API, FileInfo, Options } from "jscodeshift";
import {
	addImportSpecifiers,
	addSuggestionToSuggestionsFile,
	getAncestorBlockStatement,
	getImportAliasToRealComponentMapping,
	getSafeCodemoddedSource,
	isInClassComponent,
	removeImportSpecifiers,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1351](https://github.com/iTwin/iTwinUI/pull/1351): `toaster` import has been removed and replaced with `useToaster` which returns a toaster object with the same public methods as before.
 * > ```diff
 * > - import { toaster } from '@itwin/itwinui-react';
 * > + import { useToaster } from '@itwin/itwinui-react';
 * >
 * >   export default () => {
 * > +   const toaster = useToaster();
 * >
 * >     return <button onClick={() => toaster.positive('hello')}>show</button>
 * >   };
 * > ```
 */
const toasterToUseToaster = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	try {
		const importAliasMapping = getImportAliasToRealComponentMapping(j, root, ["toaster"]);

		// Remove imports of toaster
		const [numberOfImportSpecifiersRemoved, newHasChanged] = removeImportSpecifiers(j, root, [
			"toaster",
		]);
		hasChanged = hasChanged || newHasChanged;

		// If no import specifiers were removed, nothing to do.
		// Throw anything to skip subsequent steps
		if (numberOfImportSpecifiersRemoved === 0) {
			throw null;
		}

		// Add import of useToaster
		const [numberOfImportSpecifiersAdded, newHasChanged1] = addImportSpecifiers(j, root, [
			"useToaster",
		]);
		hasChanged = hasChanged || newHasChanged1;

		// "toaster" and its import aliases
		const allToasterVariableNames = Object.keys(importAliasMapping);

		// For each toasterVariableName, add a variable declaration for useToaster()
		allToasterVariableNames.map((toasterVariableName) => {
			// Find all uses of toasterVariableName
			const toasterVariableUses = root.find(j.Identifier, {
				name: toasterVariableName,
			});

			// For each use, find the upper-most block statement and add a variable declaration for toasterVariableName
			toasterVariableUses.forEach((path) => {
				const upperMostBlockStatement = getAncestorBlockStatement(path, "upper-most");

				// If upperMostBlockStatement is in a class component, don't add a variable declaration for toasterVariableName
				// Because the useToaster() cannot be called in a class component
				if (upperMostBlockStatement != null && isInClassComponent(upperMostBlockStatement)) {
					const message = `Could not automatically replace the usage of \`${toasterVariableName}\` with \`useToaster()\` because it is in a class component, and the \`useToaster()\` hook cannot be used in class components. Please manually do the needful change to use \`useToaster()\` instead of \`${toasterVariableName}\`. See [link](https://github.com/iTwin/iTwinUI/wiki/iTwinUI-react-v3-migration-guide#toaster)`;

					addSuggestionToSuggestionsFile({
						fileInfo,
						loc: path.node.loc,
						message,
						codemod: "toaster-to-usetoaster",
					});

					return;
				}

				const existingToasterVariableDeclaration = upperMostBlockStatement?.node.body.find(
					(node) => {
						// Atleast least one VariableDeclarator in the VariableDeclaration should have an id with name toasterVariableName
						return (
							j.VariableDeclaration.check(node) &&
							node.declarations.some(
								(declarator) =>
									j.VariableDeclarator.check(declarator) &&
									j.Identifier.check(declarator.id) &&
									declarator.id.name === toasterVariableName,
							)
						);
					},
				);

				// If there is already a variable declaration for toasterVariableName (i.e. we already added it), don't add it again
				if (!existingToasterVariableDeclaration) {
					// Create a variable called toasterVariableName and assign useToaster() to it
					const toasterVariableDeclaration = j.variableDeclaration("const", [
						j.variableDeclarator(
							j.identifier(toasterVariableName),
							j.callExpression(j.identifier("useToaster"), []),
						),
					]);
					upperMostBlockStatement?.node.body.unshift(toasterVariableDeclaration);
					hasChanged = true;
				}
			});
		});
	} finally {
		return getSafeCodemoddedSource({
			root,
			fileInfo,
			hasChanged,
		});
	}
};

export default toasterToUseToaster;
