import type { API, FileInfo, Options } from "jscodeshift";
import {
	getITwinUIReactImportDeclarations,
	getImportSpecifierName,
	getSafeCodemoddedSource,
	renameComponent,
} from "../utils/codemods";

/**
 * Codemod that partially addresses the below change:
 * > #### UserIcon / UserIconGroup
 * > - [#1383](https://github.com/iTwin/iTwinUI/pull/1383): Removed previously-deprecated `UserIcon` and `UserIconGroup` components (replaced with `Avatar` and `AvatarGroup` respectively). Also removed `userIcon` prop from `Header` (now part of `actions` prop).
 */
const userIconDeprecated = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	const renameMap = {
		UserIcon: "Avatar",
		UserIconGroup: "AvatarGroup",
	} as const;

	// ------------------------------------------------------------------------------------------------

	// Cases:
	// 1. UserIcon/UserIconGroup is imported as an alias
	// 2. UserIcon/UserIconGroup is imported as-is

	// ------------------------------------------------------------------------------------------------

	getITwinUIReactImportDeclarations(j, root).forEach((importDeclaration) => {
		importDeclaration.node.specifiers?.forEach((importSpecifier) => {
			const specifierNames = getImportSpecifierName(importSpecifier);

			// Only concerned with UserIcon/UserIconGroup
			if (!Object.keys(renameMap).includes(specifierNames.importedName ?? "")) {
				return;
			}

			// Case 1: UserIcon/UserIconGroup is imported as-is
			// Rename the component and the import specifier
			if (specifierNames.name === specifierNames.importedName) {
				if (specifierNames.name == null) {
					return;
				}

				const replacementName = renameMap[specifierNames.name as keyof typeof renameMap];

				// Rename all occurrences of the component
				const allComponents = root.findJSXElements(specifierNames.name);
				allComponents.forEach((component) => {
					const newHasChanged = renameComponent(component, replacementName);
					hasChanged = hasChanged || newHasChanged;
				});

				// Rename the import specifier
				(importSpecifier as any).local.name = replacementName;
				hasChanged = true;
			}

			// Case 2: UserIcon/UserIconGroup is imported as an alias
			// Just rename the real name but keep the alias name the same
			else {
				const replacementName =
					specifierNames.importedName != null
						? renameMap[specifierNames.importedName as keyof typeof renameMap]
						: null;

				(importSpecifier as any).imported.name = replacementName;
				hasChanged = true;
			}
		});
	});

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default userIconDeprecated;
