import { API, FileInfo, Options } from "jscodeshift";
import {
	getImportAliasToRealComponentMapping,
	getSafeCodemoddedSource,
	removeImportSpecifiers,
} from "../utils/codemods";

/**
 * Codemod for the below change:
 * > [#1265](https://github.com/iTwin/iTwinUI/pull/1265): Removed the deprecated `useTheme` hook. `<ThemeProvider>` is now always required to be wrapped around every tree within which iTwinUI components are used.
 * > ```diff
 * > + <ThemeProvider>
 * >    <App />
 * > + </ThemeProvider>
 * > ```
 * @param fileInfo
 * @param api
 * @param options
 */
const useThemeDeprecated = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	// ------------------------------------------------------------------------------------------------
	// Step 1: Find and remove the useTheme (and alias imports) calls

	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root, [
		"useTheme",
	]);

	// Iterate through each alias
	Object.entries(importAliasToRealComponentMapping).forEach(
		([useThemeAlias, realUseThemeFunctionName]) => {
			// Find all useTheme calls
			const useThemeCalls = root.find(j.CallExpression, {
				callee: {
					name: useThemeAlias,
				},
			});

			// Remove the useTheme call
			useThemeCalls.remove();
		},
	);

	// ------------------------------------------------------------------------------------------------
	// Step 2: Remove useTheme ImportSpecifier from all @itwin/itwinui-react ImportDeclarations

	const [numberOfImportSpecifiersRemoved, newHasChanged] = removeImportSpecifiers(j, root, [
		"useTheme",
	]);
	hasChanged = hasChanged || newHasChanged;

	// ------------------------------------------------------------------------------------------------

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default useThemeDeprecated;
