import type { API, Collection, FileInfo, JSCodeshift, Options, ReturnStatement } from "jscodeshift";
import {
	getSafeCodemoddedSource,
	getImportAliasToRealComponentMapping,
	getAncestorBlockStatement,
	addSuggestionToSuggestionsFile,
} from "../utils/codemods";

/**
 * Codemod to detect uses of `useToaster` outside of ThemeProvider.
 *
 * Loosely based on the below change:
 * > - [#1265](https://github.com/iTwin/iTwinUI/pull/1265): Removed the deprecated `useTheme` hook. `<ThemeProvider>` is now always required to be wrapped around every tree within which iTwinUI components are used.
 * > ```diff
 * > + <ThemeProvider>
 * >    <App />
 * > + </ThemeProvider>
 * > ```
 */
const useToasterInsideThemeProvider = (fileInfo: FileInfo, api: API, options: Options): string => {
	const j = api.jscodeshift;
	const root = j(fileInfo.source);

	let hasChanged = false;

	// ------------------------------------------------------------------------------------------------
	// Find all calls to useToaster

	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root, [
		"useToaster",
	]);

	Object.entries(importAliasToRealComponentMapping).forEach(
		([useToasterAlias, realUseToasterFunctionName]) => {
			// Find all useToaster calls
			const useToasterCalls = root.find(j.CallExpression, {
				callee: {
					name: useToasterAlias,
				},
			});

			useToasterCalls.forEach((path) => {
				const blockStatement = getAncestorBlockStatement(path);
				if (blockStatement == null) {
					return;
				}

				const returnStatement = blockStatement.node.body.find(
					(node) => node.type === "ReturnStatement",
				) as ReturnStatement | undefined;
				if (returnStatement == null) {
					return;
				}

				// If there is a ThemeProvider in scope of the return statement,
				// then there is a possibility the useToaster has no ThemeProvider ancestor.
				// So we should add a suggestion.
				if (isThemeProviderInScope(j, root, j(returnStatement))) {
					addSuggestionToSuggestionsFile({
						fileInfo,
						loc: path.node.loc,
						message:
							"Confirm that `useToaster` is scoped within at least one `ThemeProvider` ancestor. Else, the toaster may not work.",
						codemod: "usetoaster-inside-themeprovider",
					});
				}
			});
		},
	);

	return getSafeCodemoddedSource({
		root,
		fileInfo,
		hasChanged,
	});
};

export default useToasterInsideThemeProvider;

// ------------------------------------------------------------------------------------------------

const isThemeProviderInScope = (
	j: JSCodeshift,
	root: ReturnType<typeof j>,
	path: Collection<any>,
) => {
	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root, [
		"ThemeProvider",
	]);

	for (const themeProviderAlias of Object.keys(importAliasToRealComponentMapping)) {
		// Find all ThemeProvider uses within path
		const themeProviderUses = path.find(j.JSXElement, {
			openingElement: {
				name: {
					name: themeProviderAlias,
				},
			},
		});

		// return true if even one ThemeProvider alias is used
		if (themeProviderUses.length > 0) {
			return true;
		}
	}

	// No ThemeProvider alias was used
	return false;
};
