/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import {
	Alert,
	ButtonGroup,
	IconButton,
	NodeData,
	Text,
	Tooltip,
	Tree,
	TreeNode,
} from "@itwin/itwinui-react";
import "./FileTree.scss";

import React from "react";
import { MatchesContext, RepoFolderContext } from "modules/contexts/contexts";
import { SvgWindowZoomIn, SvgWindowZoomOut } from "@itwin/itwinui-icons-react";
import {
	getAutoswapMatchIndices,
	getManualMatchLineIndices,
	getMatchFileIndex,
	getSuggestionMatchLineIndices,
} from "utils/tasks";

export type FileStructure = {
	path: string;
	name: string;
	children: FileStructure[];
};

export const FileTree = () => {
	type StoryData = {
		id: string;
		label: string;
		isInMatches: boolean;
		sublabel: string;
		subItems: StoryData[];
	};

	const { folderInfo } = React.useContext(RepoFolderContext);
	const { matches, fileTree } = React.useContext(MatchesContext);
	const allNodeIds = React.useRef(new Set<string>());

	const [expandedNodes, setExpandedNodes] = React.useState<Record<string, boolean>>({});
	const onNodeExpanded = React.useCallback((nodeId: string, isExpanded: boolean) => {
		if (isExpanded) {
			setExpandedNodes((oldExpanded) => ({ ...oldExpanded, [nodeId]: true }));
		} else {
			setExpandedNodes((oldExpanded) => ({
				...oldExpanded,
				[nodeId]: false,
			}));
		}
	}, []);

	const getStoryData = React.useCallback(
		(props: FileStructure): StoryData => {
			allNodeIds.current.add(props.path);

			const subItems =
				props.children != null ? props.children.map((prop) => getStoryData(prop)) : [];

			return {
				id: props.path,
				label: props.name,
				isInMatches:
					(subItems.length == 0 && matches.map((m) => m.fileName).includes(props.path)) ||
					(subItems.length > 0 && subItems.some((item) => item.isInMatches)),
				sublabel: `Sublabel`,
				subItems: subItems.filter((item) => item.isInMatches),
			};
		},
		[matches],
	);

	const getNode = React.useCallback(
		(node: StoryData): NodeData<StoryData> => {
			return {
				subNodes: node.subItems,
				nodeId: node.id,
				node: node,
				isExpanded: expandedNodes[node.id],
				hasSubNodes: node.subItems.length > 0,
			};
		},
		[expandedNodes],
	);

	const data: StoryData[] = React.useMemo(() => {
		if (fileTree == null) {
			return [];
		}

		const storyData = getStoryData(fileTree);
		return storyData?.subItems ?? [];
	}, [getStoryData, fileTree]);

	return (
		<span className="FileTree">
			<span className="top-bar">
				<ButtonGroup>
					<Tooltip content={"Expand All"}>
						<IconButton
							onClick={() => {
								const newExpandedNodes: Record<string, boolean> = {};
								for (const nodeId of [...allNodeIds.current]) {
									newExpandedNodes[`${nodeId}`] = true;
								}
								setExpandedNodes(newExpandedNodes);
							}}
						>
							<SvgWindowZoomOut />
						</IconButton>
					</Tooltip>
					<Tooltip content={"Collapse All"}>
						<IconButton onClick={() => setExpandedNodes({})}>
							<SvgWindowZoomIn />
						</IconButton>
					</Tooltip>
				</ButtonGroup>
			</span>
			{folderInfo && (
				<Alert type="informational" className="folder-info-alert">
					<Text>
						Scanned files: <strong>{folderInfo.filesFoundCount}</strong>
						<br />
						Files with recommendations: <strong>{matches.length}</strong>
					</Text>
				</Alert>
			)}
			<Tree<StoryData>
				className="file-tree"
				data={data}
				getNode={getNode}
				nodeRenderer={React.useCallback(
					({ node, ...rest }) => {
						const isFile = node.subItems.length === 0;
						let matchFileIndex: number = -1;
						if (isFile) {
							matchFileIndex = getMatchFileIndex(matches, node.id);
						}

						const label = node.label;
						const sublabel =
							isFile &&
							matchFileIndex >= 0 &&
							`${getAutoswapMatchIndices(matches, matchFileIndex).length} | ${
								getManualMatchLineIndices(matches, matchFileIndex).length
							} | ${getSuggestionMatchLineIndices(matches, matchFileIndex).length}`;

						const treeNode = (
							<TreeNode
								className="tree-item"
								data-is-file={isFile}
								label={<Text>{label}</Text>}
								sublabel={sublabel}
								onExpanded={onNodeExpanded}
								// TODO: Add this back when TreeNode allows to pass props to the dom element
								// title={`${label}${sublabel ? `\n${sublabel}` : ""}`}
								onSelected={() => {
									if (isFile) {
										const fileCardId = `FileCard-${node.id}`;
										location.href = `#${fileCardId}`;
										document.getElementById(fileCardId)?.focus();
									}
								}}
								{...rest}
							/>
						);

						return treeNode;
					},
					[matches, onNodeExpanded],
				)}
			/>
		</span>
	);
};
