/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import React from "react";
import "./FilesSection.scss";
import { FileCard } from "./components/FileCard/FileCard";
import {
	Button,
	ButtonGroup,
	IconButton,
	ProgressRadial,
	Tab,
	Tabs,
	Text,
	Tooltip,
} from "@itwin/itwinui-react";
import { SvgExpandAll, SvgCollapseAll } from "@itwin/itwinui-icons-react";
import { MatchesContext, RepoFolderContext, RulesContext } from "modules/contexts/contexts";
import { SvgRefresh, SvgStatusError, SvgStatusSuccess } from "@itwin/itwinui-icons-react";
import { InstructionsSection } from "./components/InstructionsSection/InstructionsSection";
import { getRules } from "utils/api_calls";
import {
	getAllMatchesInRepo,
	getAllMatchLineCount,
	getIsAllChangedCount,
	getIsMatchLineMatchingFilter,
} from "utils/tasks";
import { MatchLine, MatchRule } from "data/file_types";
import { RuleCard } from "./components/RuleCard/RuleCard";

export type DisplayAggregation = "files" | "rules";

export const FilesSection = () => {
	const [displayAggregation, setDisplayAggregation] = React.useState<DisplayAggregation>("files");

	const [openedBlocksByFiles, setOpenedBlocksByFiles] = React.useState<number[]>([]);
	const [openedBlocksByRules, setOpenedBlocksByRules] = React.useState<number[]>([]);

	const [matchesGroupedByRule, setMatchesGroupedByRule] = React.useState<
		(MatchRule & {
			matchLines: (MatchLine & {
				fileName: string;
				matchFileIndex: number;
				matchLineIndex: number;
			})[];
		})[]
	>([]);

	const toggleHandler = React.useCallback(
		(isExpanding: boolean, id: number) => {
			if (displayAggregation === "files") {
				if (isExpanding) {
					setOpenedBlocksByFiles((prev) => [...prev, id]);
				} else {
					setOpenedBlocksByFiles((prev) => prev.filter((b) => b !== id));
				}
			} else {
				if (isExpanding) {
					setOpenedBlocksByRules((prev) => [...prev, id]);
				} else {
					setOpenedBlocksByRules((prev) => prev.filter((b) => b !== id));
				}
			}
		},
		[displayAggregation],
	);

	const { packagesToScan } = React.useContext(RulesContext);
	const { repoFolder, foundRepoFolder, setFoundRepoFolder, folderInfo, setFolderInfo } =
		React.useContext(RepoFolderContext);
	const {
		matches,
		setMatches,
		gettingMatches,
		setGettingMatches,
		selectedFilters,
		setSelectedFilters,
		setFileTree,
		lockAllOperations,
	} = React.useContext(MatchesContext);

	const mainButtonGroup = React.useCallback(() => {
		return (
			<ButtonGroup>
				<Tooltip content="Rescan">
					<IconButton
						onClick={async () => {
							getAllMatchesInRepo(
								repoFolder?.current ?? "",
								setFoundRepoFolder,
								setMatches,
								setFileTree,
								setFolderInfo,
								gettingMatches,
								setGettingMatches,
								packagesToScan,
							);
						}}
					>
						<SvgRefresh />
					</IconButton>
				</Tooltip>
				<Tooltip content="Expand all">
					<IconButton
						onClick={() => {
							if (displayAggregation === "files") {
								setOpenedBlocksByFiles([...Array(matches.length)].map((_, index) => index));
							} else {
								setOpenedBlocksByRules(
									[...Array(matchesGroupedByRule.length)].map((_, index) => index),
								);
							}
						}}
					>
						<SvgExpandAll />
					</IconButton>
				</Tooltip>
				<Tooltip content="Collapse all">
					<IconButton
						onClick={() => {
							if (displayAggregation === "files") {
								setOpenedBlocksByFiles([]);
							} else {
								setOpenedBlocksByRules([]);
							}
						}}
					>
						<SvgCollapseAll />
					</IconButton>
				</Tooltip>
			</ButtonGroup>
		);
	}, [
		displayAggregation,
		gettingMatches,
		matches.length,
		matchesGroupedByRule.length,
		packagesToScan,
		repoFolder,
		setFileTree,
		setFolderInfo,
		setFoundRepoFolder,
		setGettingMatches,
		setMatches,
	]);

	const filtersButtonGroup = React.useCallback(() => {
		const genericButton = ({
			filter,
			filterCount,
			startIcon,
			customTooltipContent,
			customButtonContent,
		}: {
			filter: MatchLine["type"] | "completed";
			filterCount: number;
			startIcon: JSX.Element;
			customTooltipContent?: (isActive: boolean) => string;
			customButtonContent?: React.ReactNode;
		}) => {
			const onButtonClick = (filter: MatchLine["type"] | "completed") => {
				setSelectedFilters((prev: (MatchLine["type"] | "completed")[]) => {
					if (prev.includes(filter)) {
						return prev.filter((f) => f !== filter);
					} else {
						return [...prev, filter];
					}
				});
			};
			const isActive = selectedFilters.includes(filter);

			return (
				<Tooltip
					content={
						customTooltipContent
							? customTooltipContent(isActive)
							: `${isActive ? "Hide" : "Show"} ${filter}`
					}
				>
					<Button
						startIcon={startIcon}
						onClick={() => onButtonClick(filter)}
						data-active={isActive}
						className="button"
					>
						{customButtonContent ? customButtonContent : filter} <strong>{filterCount}</strong>
					</Button>
				</Tooltip>
			);
		};

		const typeButton = (type: MatchLine["type"]) => {
			return genericButton({
				filter: type,
				filterCount: getAllMatchLineCount(matches, type),
				startIcon: <SvgStatusError className="status-icon" data-error={`${type}`} />,
			});
		};

		const completedButton = () => {
			return genericButton({
				filter: "completed",
				filterCount: getIsAllChangedCount(matches),
				startIcon: <SvgStatusSuccess className="status-icon" data-success={true} />,
				customTooltipContent: (isActive) => (isActive ? "Show completed" : "Hide completed"),
				customButtonContent: `Auto-hide completed`,
			});
		};

		return (
			<>
				<ButtonGroup className="filters-button-group">
					{typeButton("autoswap")}
					{typeButton("manual")}
					{typeButton("suggestion")}
				</ButtonGroup>
				<ButtonGroup className="filters-button-group">{completedButton()}</ButtonGroup>
			</>
		);
	}, [matches, selectedFilters, setSelectedFilters]);

	const filterBar = React.useCallback(() => {
		return (
			<span className="filters">
				{mainButtonGroup()}
				{filtersButtonGroup()}
			</span>
		);
	}, [filtersButtonGroup, mainButtonGroup]);

	/**
	 * Calls `setMatchesGroupedByRule(rules)`. Here, `rules` contains the rule and the `lineMatches` for each rule.
	 */
	const getCardByRules = React.useCallback(async () => {
		if (displayAggregation !== "rules") {
			return;
		}

		const rules:
			| (MatchRule & {
					matchLines?: (MatchLine & {
						fileName: string;
						matchFileIndex: number;
						matchLineIndex: number;
					})[];
			  })[]
			| null = await getRules(packagesToScan);
		if (rules == null) {
			return;
		}

		for (let i = 0; i < matches.length; i++) {
			const matchFile = matches[i];
			for (let j = 0; j < matchFile.matchLines.length; j++) {
				const matchLine = matchFile.matchLines[j];

				// Assuming "replace" in all rules are unique
				const ruleIndex = rules.findIndex((rule) => rule.replace === matchLine.replaceRule);

				if (ruleIndex < 0) {
					continue;
				}

				const matchLineToPush = {
					...matchLine,
					fileName: matchFile.fileName,
					matchFileIndex: i,
					matchLineIndex: j,
				};
				if ("matchLines" in rules[ruleIndex]) {
					rules[ruleIndex].matchLines?.push(matchLineToPush);
				} else {
					rules[ruleIndex].matchLines = [matchLineToPush];
				}
			}
		}

		const rulesWithMatchLines = rules.filter((rule) => "matchLines" in rule) as (MatchRule & {
			matchLines: (MatchLine & {
				fileName: string;
				matchFileIndex: number;
				matchLineIndex: number;
			})[];
		})[];

		setMatchesGroupedByRule(rulesWithMatchLines);
	}, [displayAggregation, matches, packagesToScan]);

	React.useEffect(() => {
		getCardByRules();
	}, [getCardByRules, matches, displayAggregation]);

	const getIsSomeMatchLinesMatchingFilter = React.useCallback(
		(matchLines: MatchLine[]) => {
			return matchLines.some((ml) => getIsMatchLineMatchingFilter(ml, selectedFilters));
		},
		[selectedFilters],
	);

	const cardsByFiles = React.useCallback(() => {
		return matches.map((matchFile, index) => {
			const isSomeMatchLinesMatchingFilter = getIsSomeMatchLinesMatchingFilter(
				matchFile.matchLines,
			);
			if (isSomeMatchLinesMatchingFilter) {
				return (
					<FileCard
						key={index}
						matchFileIndex={index}
						isExpanded={openedBlocksByFiles.includes(index)}
						onToggle={(isExpanding) => toggleHandler(isExpanding, index)}
					/>
				);
			}
		});
	}, [getIsSomeMatchLinesMatchingFilter, matches, openedBlocksByFiles, toggleHandler]);

	const cardsByRules = React.useCallback(() => {
		return matchesGroupedByRule.map((matchRule, index) => {
			const isSomeMatchLinesMatchingFilter = getIsSomeMatchLinesMatchingFilter(
				matchRule.matchLines,
			);

			if (isSomeMatchLinesMatchingFilter) {
				return (
					<RuleCard
						key={index}
						matchesGroupedByRule={matchRule}
						isExpanded={openedBlocksByRules.includes(index)}
						onToggle={(isExpanding) => toggleHandler(isExpanding, index)}
					/>
				);
			}
		});
	}, [getIsSomeMatchLinesMatchingFilter, matchesGroupedByRule, openedBlocksByRules, toggleHandler]);

	const cardsSection = React.useCallback(() => {
		const cards = displayAggregation === "files" ? cardsByFiles() : cardsByRules();

		const getNonCardsContentText = () => {
			if (foundRepoFolder) {
				if (matches.length === 0) {
					return "No matches found. You are likely completely migrated to the new version!";
				}
				if (cards.every((c) => c == null)) {
					return "No matches match selected filters. Try removing filters.";
				}
			} else {
				return "Enter repo path and press search to view matches";
			}

			return null;
		};

		const nonCardsContentText = getNonCardsContentText();
		const content = nonCardsContentText ? <Text>{nonCardsContentText}</Text> : cards;

		return (
			<span className="cards">
				{gettingMatches ? (
					<span>
						<ProgressRadial indeterminate />
						<Text>
							Scanning {folderInfo != null ? folderInfo.filesFoundCount : ""} files for matches.
							This could take a while...
						</Text>
					</span>
				) : (
					<>{content}</>
				)}
			</span>
		);
	}, [
		cardsByFiles,
		cardsByRules,
		displayAggregation,
		folderInfo,
		foundRepoFolder,
		gettingMatches,
		matches.length,
	]);

	return (
		<span className="FilesSection" data-lock-all-operations={lockAllOperations}>
			{filterBar()}
			<InstructionsSection />
			<Tabs
				activeIndex={displayAggregation === "files" ? 0 : 1}
				labels={[<Tab key={1} label="By files" />, <Tab key={2} label="By rules" />]}
				onTabSelected={(index) => setDisplayAggregation(index === 0 ? "files" : "rules")}
				wrapperClassName="tabs-wrapper"
				tabsClassName="tabs"
				contentClassName="tabs-content"
			>
				{cardsSection()}
			</Tabs>
		</span>
	);
};
