/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import {
	ExpandableBlock,
	MenuItem,
	Text,
	Tooltip,
	ProgressRadial,
	IconButton,
	SplitButton,
} from "@itwin/itwinui-react";
import "./FileCard.scss";
import { SvgStatusSuccess, SvgStatusError } from "@itwin/itwinui-icons-react";
import { LineChange } from "./components/LineChange/LineChange";
import React from "react";
import { MatchesContext, RepoFolderContext, RulesContext } from "modules/contexts/contexts";
import {
	displayToast,
	getAutoswapMatchIndices,
	getIsAllAutoswapChangedInFile,
	getIsAllChangedInFile,
	getIsAllManualChangedInFile,
	getIsAllSuggestionChangedInFile,
	getIsMatchLineMatchingFilter,
	getManualMatchLines,
	getSuggestionMatchLines,
	getAutoswapChangesInFile,
	toggleAllManualInFile,
	toggleAllNonAutoswapInFile,
	toggleAllSuggestionInFile,
} from "utils/tasks";
import { SvgRefresh } from "@itwin/itwinui-icons-react";
import { getMatchesInFile } from "utils/api_calls";
import { MatchFile, MatchLine, ToastMessage } from "data/file_types";
import { getFileNameAndPath } from "utils/utils";

export type FileCardProps = {
	matchFileIndex: number;
	isExpanded: boolean;
	onToggle: (isExpanding: boolean) => void;
};

export const FileCard = (props: FileCardProps) => {
	const { matchFileIndex, isExpanded, onToggle } = props;

	const { packagesToScan } = React.useContext(RulesContext);
	const { repoFolder } = React.useContext(RepoFolderContext);
	const { matches, setMatches, selectedFilters, gettingMatches } = React.useContext(MatchesContext);
	const file = matches[matchFileIndex];

	const autoswapMatchLinesIndices = getAutoswapMatchIndices(matches, matchFileIndex);
	const manualMatchLines = getManualMatchLines(matches, matchFileIndex);
	const suggestionMatchLines = getSuggestionMatchLines(matches, matchFileIndex);

	const isChanged = getIsAllChangedInFile(matches, matchFileIndex);
	const [isChanging, setIsChanging] = React.useState(false);

	const isAllAutoswapChangedInFile = getIsAllAutoswapChangedInFile(matches, matchFileIndex);
	const isAllManualChangedInFile = getIsAllManualChangedInFile(matches, matchFileIndex);
	const isAllSuggestionChangedInFile = getIsAllSuggestionChangedInFile(matches, matchFileIndex);

	const myRef = React.useRef<any>(null);
	const fileNameAndPath = React.useMemo(
		() => getFileNameAndPath(file.fileName, repoFolder?.current ?? ""),
		[file.fileName, repoFolder],
	);

	const getMenuItemText = (isAllChanged: boolean, type: MatchLine["type"], count: number) => {
		const implementOrUndo = count === 0 || !isAllChanged ? "Implement" : "Undo";
		return `${implementOrUndo} all "${type}" in file (${count})`;
	};

	const buttonMenuItems = React.useCallback(
		(close: () => void) => [
			<MenuItem
				key={1}
				onClickCapture={async (e) => {
					e.stopPropagation();

					close();
					let newMatches = JSON.parse(JSON.stringify(matches));
					newMatches = await getAutoswapChangesInFile(
						newMatches,
						matchFileIndex,
						autoswapMatchLinesIndices,
						isAllAutoswapChangedInFile,
					);
					setMatches(newMatches);
				}}
				disabled={autoswapMatchLinesIndices.length <= 0}
			>
				{getMenuItemText(isAllAutoswapChangedInFile, "autoswap", autoswapMatchLinesIndices.length)}
			</MenuItem>,
			<MenuItem
				key={2}
				onClickCapture={(e) => {
					e.stopPropagation();

					close();
					toggleAllManualInFile(matches, setMatches, matchFileIndex);
				}}
				disabled={manualMatchLines.length <= 0}
			>
				{getMenuItemText(isAllManualChangedInFile, "manual", manualMatchLines.length)}
			</MenuItem>,
			<MenuItem
				key={3}
				onClickCapture={(e) => {
					e.stopPropagation();

					close();
					console.log("clicked");
					toggleAllSuggestionInFile(matches, setMatches, matchFileIndex);
				}}
				disabled={suggestionMatchLines.length <= 0}
			>
				{getMenuItemText(isAllSuggestionChangedInFile, "suggestion", suggestionMatchLines.length)}
			</MenuItem>,
		],
		[
			autoswapMatchLinesIndices,
			isAllAutoswapChangedInFile,
			isAllManualChangedInFile,
			isAllSuggestionChangedInFile,
			manualMatchLines.length,
			matchFileIndex,
			matches,
			setMatches,
			suggestionMatchLines.length,
		],
	);

	const iconButtonClick = React.useCallback(async () => {
		const autoswapMatchLineIndices = getAutoswapMatchIndices(matches, matchFileIndex);
		setIsChanging(true);

		// Non autoswap
		toggleAllNonAutoswapInFile(matches, setMatches, matchFileIndex);

		// Autoswap
		const newMatches = await getAutoswapChangesInFile(
			matches,
			matchFileIndex,
			autoswapMatchLineIndices,
			isChanged,
		);
		setMatches(newMatches);

		setIsChanging(false);
	}, [isChanged, matchFileIndex, matches, setMatches]);

	const statusIcon = React.useCallback(
		() => (
			<Tooltip content={isChanged ? "Undo all change(s)" : "Make all changes(s)"}>
				<SplitButton
					className="split-icon-btn"
					menuItems={buttonMenuItems}
					size="small"
					disabled={isChanging}
					onClick={(e) => {
						e.stopPropagation();
						iconButtonClick();
					}}
					menuButtonProps={{
						onClick: (e) => {
							e.stopPropagation();
						},
					}}
				>
					{isChanging ? (
						<ProgressRadial className="progress-indicator" indeterminate />
					) : isChanged ? (
						<SvgStatusSuccess className="status-icon" data-success="true" aria-hidden />
					) : (
						<SvgStatusError className={`status-icon`} aria-hidden />
					)}
				</SplitButton>
			</Tooltip>
		),
		[buttonMenuItems, iconButtonClick, isChanged, isChanging],
	);

	const headerContent = React.useCallback(() => {
		const subTitle = `${autoswapMatchLinesIndices.length} autoswap ∙ ${manualMatchLines.length} manual ∙ ${suggestionMatchLines.length} suggestion`;

		return (
			<span className="content">
				<span
					className="title-path"
					title={`${fileNameAndPath.fileName} (${fileNameAndPath.path})`}
				>
					<Text className={"title"} variant="leading" as="b">
						{fileNameAndPath.fileName}
					</Text>
					<Text className={"path"} variant="small" as="b">
						{fileNameAndPath.path}
					</Text>
				</span>

				<Text className={"sub-title"} variant="small" title={subTitle}>
					{subTitle}
				</Text>
			</span>
		);
	}, [
		autoswapMatchLinesIndices.length,
		fileNameAndPath.fileName,
		fileNameAndPath.path,
		manualMatchLines.length,
		suggestionMatchLines.length,
	]);

	const rescanFileButton = React.useCallback(() => {
		const onClick = async (e: React.MouseEvent<HTMLButtonElement>) => {
			e.stopPropagation();
			const matchFiles = await getMatchesInFile(gettingMatches, file.fileName, packagesToScan);

			let toastMessage: ToastMessage | null | undefined;

			setMatches((prev: MatchFile[]) => {
				const newMatches = JSON.parse(JSON.stringify(prev));
				const fileName = getFileNameAndPath(
					matches[matchFileIndex].fileName,
					repoFolder?.current ?? "",
				).fileName;

				if (matchFiles != null) {
					if (matchFiles.length === 1) {
						const matchFile = matchFiles[0];
						// Replace
						newMatches[matchFileIndex] = matchFile;
						toastMessage = {
							type: "positive",
							content: `Rescanned: ${fileName}`,
						};
					} else {
						// Remove
						newMatches.splice(matchFileIndex, 1); // Remove file from matches
						toastMessage = {
							type: "positive",
							content: `Rescanned: Since all recommendations are complete, removing ${fileName} from the list`,
						};
					}
				} else {
					// error; do nothing; show toast
					toastMessage = {
						type: "negative",
						content: "Error rescanning file search",
					};
				}
				return newMatches;
			});

			if (toastMessage != null) {
				displayToast(toastMessage.type, toastMessage.content);
			}
		};

		return (
			<Tooltip content="Rescan file">
				<IconButton className="refresh-btn" size="small" onClick={onClick}>
					<SvgRefresh />
				</IconButton>
			</Tooltip>
		);
	}, [
		file.fileName,
		gettingMatches,
		matchFileIndex,
		matches,
		packagesToScan,
		repoFolder,
		setMatches,
	]);

	return (
		<ExpandableBlock
			id={`FileCard-${file.fileName}`}
			title={
				<span className="header" ref={myRef}>
					<span className="left">
						{statusIcon()}
						{headerContent()}
					</span>
					<span className="right">{rescanFileButton()}</span>
				</span>
			}
			className="FileCard"
			isExpanded={isExpanded}
			onToggle={(isExpanding) => {
				onToggle(isExpanding);
			}}
		>
			<div className="content">
				{file.matchLines.map((matchLine, index) => {
					if (getIsMatchLineMatchingFilter(matchLine, selectedFilters)) {
						return (
							<LineChange
								key={index}
								fileName={file.fileName}
								lineChangeByFileProps={{
									matchFileIndex: matchFileIndex,
									matchLineIndex: index,
								}}
							/>
						);
					}
					return null;
				})}
			</div>
		</ExpandableBlock>
	);
};
