/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import { Code, Text, IconButton, ProgressRadial, Tooltip, Anchor } from "@itwin/itwinui-react";
import { SvgStatusSuccess, SvgStatusError } from "@itwin/itwinui-icons-react";
import "./LineChange.scss";
import React from "react";
import { MatchFile, MatchLine } from "data/file_types";
import { MatchesContext, RepoFolderContext } from "modules/contexts/contexts";
import { displayToast, getAutoswapChangesInFile, toggleNonAutoswapInFile } from "utils/tasks";
import { getFileNameAndPath } from "utils/utils";
import { CustomReactMarkdown } from "app/_ui/CustomReactMarkdown";

export type LineChangeByFileProps = {
	matchFileIndex: number;
	matchLineIndex: number;
};

export type LineChangeByRuleProps = {
	matchLine: MatchLine;
	matchFileIndex: number;
	matchLineIndex: number;
};

export type LineChangeProps = {
	fileName: string;
	lineChangeByFileProps?: LineChangeByFileProps;
	lineChangeByRuleProps?: LineChangeByRuleProps;
};

export const LineChange = (props: LineChangeProps) => {
	const { fileName, lineChangeByFileProps, lineChangeByRuleProps } = props;

	const { repoFolder } = React.useContext(RepoFolderContext);
	const { matches, setMatches } = React.useContext(MatchesContext);

	const [isChanging, setIsChanging] = React.useState(false);
	const matchLine =
		lineChangeByFileProps != null
			? matches[lineChangeByFileProps.matchFileIndex].matchLines[
					lineChangeByFileProps.matchLineIndex
			  ]
			: lineChangeByRuleProps!.matchLine;

	const fileNameAndPath = React.useMemo(
		() => getFileNameAndPath(fileName, repoFolder?.current ?? ""),
		[fileName, repoFolder],
	);

	const statusIcon = React.useCallback(() => {
		const statusIconTemplate = ({ onClick }: { onClick: () => void }) => {
			return (
				<Tooltip content={matchLine.changed ? "Undo change" : "Make change"}>
					<IconButton size="small" onClick={onClick} disabled={isChanging}>
						{isChanging ? (
							<ProgressRadial className="progress-indicator" indeterminate />
						) : matchLine.changed ? (
							<SvgStatusSuccess className="status-icon" data-success="true" aria-hidden />
						) : (
							<SvgStatusError
								className={`status-icon`}
								data-error={`${matchLine.type}`}
								aria-hidden
							/>
						)}
					</IconButton>
				</Tooltip>
			);
		};

		if (lineChangeByFileProps != null) {
			return statusIconTemplate({
				onClick: async () => {
					setIsChanging(true);
					if (matchLine.type === "autoswap") {
						const newMatches = JSON.parse(JSON.stringify(matches));
						await getAutoswapChangesInFile(
							newMatches,
							lineChangeByFileProps.matchFileIndex,
							[lineChangeByFileProps.matchLineIndex],
							matchLine.changed,
						);
						setMatches(newMatches);
					} else {
						toggleNonAutoswapInFile(
							setMatches,
							lineChangeByFileProps.matchFileIndex,
							lineChangeByFileProps.matchLineIndex,
							!matches[lineChangeByFileProps.matchFileIndex].matchLines[
								lineChangeByFileProps.matchLineIndex
							].changed,
						);
					}
					setIsChanging(false);
				},
			});
		} else if (lineChangeByRuleProps != null) {
			return statusIconTemplate({
				onClick: async () => {
					setIsChanging(true);
					if (matchLine.type === "autoswap") {
						let newMatches = JSON.parse(JSON.stringify(matches));
						await getAutoswapChangesInFile(
							newMatches,
							lineChangeByRuleProps.matchFileIndex,
							[lineChangeByRuleProps.matchLineIndex],
							matchLine.changed,
						);
						setMatches(newMatches);
					} else {
						setMatches((prev: MatchFile[]) => {
							const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
							newMatches[lineChangeByRuleProps.matchFileIndex].matchLines[
								lineChangeByRuleProps.matchLineIndex
							].changed =
								!newMatches[lineChangeByRuleProps.matchFileIndex].matchLines[
									lineChangeByRuleProps.matchLineIndex
								].changed;
							return newMatches;
						});
					}
					setIsChanging(false);
				},
			});
		}

		return null;
	}, [
		isChanging,
		lineChangeByFileProps,
		lineChangeByRuleProps,
		matchLine.changed,
		matchLine.type,
		matches,
		setMatches,
	]);

	const getRemoveLine = React.useCallback(() => {
		return (
			<Code className="line-remove">
				{matchLine.line.substring(0, matchLine.line.indexOf(matchLine.replace))}
				<strong className="code-remove">{matchLine.replace}</strong>
				{matchLine.line.substring(
					matchLine.line.indexOf(matchLine.replace) + matchLine.replace.length,
				)}
			</Code>
		);
	}, [matchLine.line, matchLine.replace]);

	const getAddLine = React.useCallback(() => {
		return (
			<Code className="line-add">
				{matchLine.line.substring(0, matchLine.line.indexOf(matchLine.replace))}
				<strong className="code-add">{matchLine.with}</strong>
				{matchLine.line.substring(
					matchLine.line.indexOf(matchLine.replace) + matchLine.replace.length,
				)}
			</Code>
		);
	}, [matchLine.line, matchLine.replace, matchLine.with]);

	const lineNumberLink = React.useCallback(() => {
		const fileNameWithLineAndColumn = `${fileName}:${matchLine.line_number}:${matchLine.columnNumber}`;

		return (
			<Tooltip content={"Open"}>
				<Anchor
					className="line-number-link"
					href={`file://${fileNameWithLineAndColumn}`}
					onClick={async (e) => {
						e.preventDefault();

						displayToast(
							"informational",
							`Opening ${fileNameAndPath.fileName} using default system application`,
						);

						const queryParams = `fileName=${fileName}&lineNumber=${matchLine.line_number}&columnNumber=${matchLine.columnNumber}`;
						const response = await fetch(`/api/open_file?${queryParams}`, {
							method: "GET",
							headers: { "Content-Type": "application/json" },
						});

						if (!response.ok) {
							displayToast("negative", `Error opening ${fileNameAndPath.fileName}`);
						}
					}}
				>
					<Text className="text">L{matchLine.line_number}</Text>
				</Anchor>
			</Tooltip>
		);
	}, [fileName, fileNameAndPath.fileName, matchLine.columnNumber, matchLine.line_number]);

	const descriptionText = React.useMemo(
		() => (lineChangeByFileProps != null ? matchLine.description ?? "" : ""),
		[lineChangeByFileProps, matchLine.description],
	);

	return (
		<span className="LineChange">
			<span className="content">
				{statusIcon()}
				{lineNumberLink()}
				<span className="main">
					{lineChangeByFileProps != null ? (
						<>
							<Text className="type">({matchLine.type}) </Text>
							<span title={descriptionText}>
								<CustomReactMarkdown className="description">{descriptionText}</CustomReactMarkdown>
							</span>
						</>
					) : (
						lineChangeByRuleProps != null && (
							<span
								className="file-name-path"
								title={`${fileNameAndPath.fileName} (${fileNameAndPath.path})`}
							>
								<Text className="file-name">{fileNameAndPath.fileName}</Text>
								<Text className="path">{fileNameAndPath.path}</Text>
							</span>
						)
					)}
				</span>
			</span>

			<span className="code">
				{getRemoveLine()}
				{(matchLine.type === "autoswap" || matchLine.with.length > 0) && getAddLine()}
			</span>
		</span>
	);
};
