import { SvgStatusSuccess, SvgUndo } from "@itwin/itwinui-icons-react";
import {
	Button,
	ButtonGroup,
	Checkbox,
	Dialog,
	Fieldset,
	IconButton,
	InputGroup,
	Text,
	Tooltip,
} from "@itwin/itwinui-react";
import React from "react";
import { iTwinPackage, MatchLine } from "data/file_types";
import { MatchesContext, RepoFolderContext, RulesContext } from "modules/contexts/contexts";
import { displayToast, getAllMatchesInRepo } from "utils/tasks";
import "./InstructionsSection.scss";
import {
	getAllAutoswapChangesDialogContent,
	getAllManualChangesDialogContent,
	getAllSuggestionChangesDialogContent,
	getUndoAllAutoswapChangesDialogContent,
	getUndoAllManualChangesDialogContent,
	getUndoAllSuggestionChangesDialogContent,
} from "./utils/dialogContent";
import { CustomReactMarkdown } from "app/_ui/CustomReactMarkdown";

export type DialogContent = {
	type: MatchLine["type"];
	title: JSX.Element;
	content: JSX.Element;
	proceedButtonClickHandler: () => void;
};

export const InstructionsSection = () => {
	const { repoFolder, setFoundRepoFolder, setFolderInfo } = React.useContext(RepoFolderContext);
	const {
		matches,
		setMatches,
		gettingMatches,
		setGettingMatches,
		setFileTree,
		lockAllOperations,
		setLockAllOperations,
	} = React.useContext(MatchesContext);
	const { packagesToScan, setPackagesToScan } = React.useContext(RulesContext);

	const [isOpen, setIsOpen] = React.useState(false);

	const [dialogContent, setDialogContent] = React.useState<DialogContent | null>(null);

	const closeDialog = () => {
		setIsOpen(false);
	};

	const onClose = React.useCallback(() => {
		closeDialog();
	}, []);

	const secondaryButtonHandle = React.useCallback(() => {
		closeDialog();
	}, []);

	// Open the dialog when dialog content is changed
	React.useEffect(() => {
		setIsOpen(true);
	}, [dialogContent]);

	const categoriesOfRecommendations = React.useCallback(() => {
		const implementAllButton = (type: MatchLine["type"]) => {
			const implementAllTooltipContent =
				type === "autoswap" ? `Implement ALL ${type}` : `Mark ALL ${type} as done`;
			const undoAllTooltipContent = `Undo ALL ${type}`;

			const implementOnClick = () =>
				setDialogContent(
					type === "autoswap"
						? getAllAutoswapChangesDialogContent(
								matches,
								setMatches,
								lockAllOperations,
								setLockAllOperations,
						  )
						: type === "manual"
						? getAllManualChangesDialogContent(matches, setMatches)
						: getAllSuggestionChangesDialogContent(matches, setMatches),
				);

			const undoOnClick = () =>
				setDialogContent(
					type === "autoswap"
						? getUndoAllAutoswapChangesDialogContent(
								matches,
								setMatches,
								lockAllOperations,
								setLockAllOperations,
						  )
						: type === "manual"
						? getUndoAllManualChangesDialogContent(matches, setMatches)
						: getUndoAllSuggestionChangesDialogContent(matches, setMatches),
				);

			return (
				<ButtonGroup>
					<Tooltip content={implementAllTooltipContent}>
						<IconButton size="small" onClick={() => implementOnClick()}>
							<SvgStatusSuccess className="status-icon" data-error={type} />
						</IconButton>
					</Tooltip>
					<Tooltip content={undoAllTooltipContent}>
						<IconButton size="small" onClick={() => undoOnClick()}>
							<SvgUndo className="status-icon" data-error={type} />
						</IconButton>
					</Tooltip>
				</ButtonGroup>
			);
		};

		return (
			<ul className="categories-of-recommendations">
				<li>
					<span className="icon-li">
						{implementAllButton("autoswap")}
						<Text>
							{`"autoswap": Changes that the tool can `}
							<strong>automatically</strong>
							{` implement (nearly perfect
              replacements).`}
						</Text>
					</span>
				</li>
				<li>
					<span className="icon-li">
						{implementAllButton("manual")}
						<Text>
							{`"manual": Changes that the tool can recommend but has to done `}
							<strong>manually</strong>
						</Text>
					</span>
				</li>
				<li>
					<span className="icon-li">
						{implementAllButton("suggestion")}
						<Text>
							{`"suggestion": `}
							<strong>Notes</strong>
							{` about possible issues/needed
              replacements`}
						</Text>
					</span>
				</li>
			</ul>
		);
	}, [lockAllOperations, matches, setLockAllOperations, setMatches]);

	const left = React.useCallback(() => {
		const instructions = `* Enter repo path into the text input above and press "Search".
    * gitignored and untracked files will be ignored
* The tool then scans the path for migration suggestions. It might take a while.
    * If it takes too long:
      * Pass a git repo or a sub-folder of a git repo as the repo path and then ignore files/folders by adding it to \`.gitignore\`
      * Run the tool multiple times with each smaller sub-folder`;

		return (
			<span className="left">
				<Text variant="subheading">Instructions:</Text>
				<CustomReactMarkdown>{instructions}</CustomReactMarkdown>
				<Text variant="subheading">Categories of recommendations:</Text>
				<Text>Click the implement/undo buttons to implement/undo all changes at once.</Text>
				{categoriesOfRecommendations()}
			</span>
		);
	}, [categoriesOfRecommendations]);

	const right = React.useCallback(() => {
		return (
			<span className="right">
				<Fieldset
					legend="Scan options"
					style={{ display: "flex", flexDirection: "column", gap: 11 }}
				>
					<InputGroup
						label={"Which rules to scan for?"}
						message={"Selecting more could increase scan time"}
					>
						<Checkbox
							label={"iTwinUI-variables: v0 > v1"}
							checked={packagesToScan.includes("itwinui-variables")}
							onChange={(event) =>
								setPackagesToScan((prev: iTwinPackage[]): iTwinPackage[] =>
									event.target.checked
										? [...prev, "itwinui-variables"]
										: prev.filter((pkg) => pkg !== "itwinui-variables"),
								)
							}
						/>
						<Checkbox
							label={"iTwinUI-css: v0 > v1"}
							checked={packagesToScan.includes("itwinui-css")}
							onChange={(event) =>
								setPackagesToScan((prev: iTwinPackage[]): iTwinPackage[] =>
									event.target.checked
										? [...prev, "itwinui-css"]
										: prev.filter((pkg) => pkg !== "itwinui-css"),
								)
							}
						/>
						<Checkbox
							label={"iTwinUI-React: v1 > v2 (breaking changes)"}
							checked={packagesToScan.includes("itwinui-react")}
							onChange={(event) =>
								setPackagesToScan((prev: iTwinPackage[]): iTwinPackage[] =>
									event.target.checked
										? [...prev, "itwinui-react"]
										: prev.filter((pkg) => pkg !== "itwinui-react"),
								)
							}
						/>
					</InputGroup>
					<Button
						styleType="high-visibility"
						onClick={() => {
							if (packagesToScan.length === 0) {
								displayToast("negative", "Choose at least one rule set");
							} else {
								getAllMatchesInRepo(
									repoFolder?.current ?? "",
									setFoundRepoFolder,
									setMatches,
									setFileTree,
									setFolderInfo,
									gettingMatches,
									setGettingMatches,
									packagesToScan,
								);
							}
						}}
					>
						Save changes
					</Button>
				</Fieldset>
			</span>
		);
	}, [
		gettingMatches,
		packagesToScan,
		repoFolder,
		setFileTree,
		setFolderInfo,
		setFoundRepoFolder,
		setGettingMatches,
		setMatches,
		setPackagesToScan,
	]);

	const dialog = React.useCallback(
		(dialogContent: DialogContent | null) => {
			if (dialogContent == null) {
				return null;
			}

			const { type, title, content, proceedButtonClickHandler } = dialogContent;

			return (
				<Dialog
					className="dialog-wrapper"
					isOpen={isOpen}
					onClose={onClose}
					closeOnEsc
					closeOnExternalClick
					preventDocumentScroll
					trapFocus
					setFocus
					isDismissible
				>
					<Dialog.Backdrop />
					<Dialog.Main className="dialog">
						<Dialog.TitleBar className="dialog-titlebar" titleText={title} />
						<Dialog.Content>{content}</Dialog.Content>
						<Dialog.ButtonBar className="dialog-button-bar">
							<Button
								styleType="cta"
								onClick={() => {
									proceedButtonClickHandler();
									closeDialog();
								}}
								className="proceed-button"
								data-error={type}
							>
								Proceed
							</Button>
							<Button onClick={secondaryButtonHandle}>Cancel</Button>
						</Dialog.ButtonBar>
					</Dialog.Main>
				</Dialog>
			);
		},
		[isOpen, onClose, secondaryButtonHandle],
	);

	return (
		<span className="InstructionsSection">
			<span className="content">
				{left()}
				{right()}
			</span>

			{dialog(dialogContent)}
		</span>
	);
};
