/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import { ExpandableBlock, Text, Tooltip, ProgressRadial, Button } from "@itwin/itwinui-react";
import "../FileCard/FileCard.scss";
import { SvgStatusSuccess, SvgStatusError } from "@itwin/itwinui-icons-react";
import React from "react";
import { MatchesContext } from "modules/contexts/contexts";
import { getIsMatchLineMatchingFilter, toggleAllChangesInRule } from "utils/tasks";
import { MatchLine, MatchRule } from "data/file_types";
import { LineChange } from "../FileCard/components/LineChange/LineChange";
import { CustomReactMarkdown } from "app/_ui/CustomReactMarkdown";

export type RuleCardProps = {
	matchesGroupedByRule: MatchRule & {
		matchLines: (MatchLine & {
			fileName: string;
			matchFileIndex: number;
			matchLineIndex: number;
		})[];
	};
	isExpanded: boolean;
	onToggle: (isExpanding: boolean) => void;
};

export const RuleCard = (props: RuleCardProps) => {
	const { matchesGroupedByRule, isExpanded, onToggle } = props;

	const { matches, setMatches, selectedFilters } = React.useContext(MatchesContext);

	const isChanged =
		matchesGroupedByRule.matchLines.filter((ml) => ml.changed).length ===
		matchesGroupedByRule.matchLines.length;
	const [isChanging, setIsChanging] = React.useState(false);

	const iconButtonClick = React.useCallback(async () => {
		setIsChanging(true);
		await toggleAllChangesInRule(
			matchesGroupedByRule,
			matchesGroupedByRule.matchLines,
			matches,
			setMatches,
		);
		setIsChanging(false);
	}, [matches, matchesGroupedByRule, setMatches]);

	const statusIcon = React.useCallback(
		() => (
			<Tooltip content={isChanged ? "Undo all change(s)" : "Make all changes(s)"}>
				<Button
					className="split-icon-btn"
					size="small"
					disabled={isChanging}
					onClick={(e) => {
						e.stopPropagation();
						iconButtonClick();
					}}
				>
					{isChanging ? (
						<ProgressRadial indeterminate />
					) : isChanged ? (
						<SvgStatusSuccess className="status-icon" data-success="true" aria-hidden />
					) : (
						<SvgStatusError className={`status-icon`} aria-hidden />
					)}
				</Button>
			</Tooltip>
		),
		[iconButtonClick, isChanged, isChanging],
	);

	const headerContent = React.useCallback(() => {
		const typeText = () => (
			<Text className="path" variant="small" data-error={matchesGroupedByRule.type}>
				{matchesGroupedByRule.type} <strong>{matchesGroupedByRule.matchLines.length}</strong>
			</Text>
		);

		return (
			<span className="content">
				<span className="title-path">
					<Text className={"title"} variant="leading">
						<CustomReactMarkdown>{`\`${matchesGroupedByRule.replace}\` > \`${matchesGroupedByRule.with}\``}</CustomReactMarkdown>
					</Text>
					{typeText()}
				</span>

				<Text className={"sub-title"} variant="small" as="p">
					<CustomReactMarkdown>{matchesGroupedByRule.description}</CustomReactMarkdown>
				</Text>
			</span>
		);
	}, [
		matchesGroupedByRule.description,
		matchesGroupedByRule.matchLines.length,
		matchesGroupedByRule.replace,
		matchesGroupedByRule.type,
		matchesGroupedByRule.with,
	]);

	const myRef = React.useRef<any>(null);

	return (
		<ExpandableBlock
			title={
				<span className="header" ref={myRef}>
					<span className="left">
						{statusIcon()}
						{headerContent()}
					</span>
				</span>
			}
			className="FileCard"
			isExpanded={isExpanded}
			onToggle={(isExpanding) => {
				onToggle(isExpanding);
			}}
		>
			{matchesGroupedByRule.matchLines.map((matchLine, index) => {
				if (getIsMatchLineMatchingFilter(matchLine, selectedFilters)) {
					return (
						<LineChange
							key={index}
							fileName={matchLine.fileName}
							lineChangeByRuleProps={{
								matchLine: matchLine,
								matchFileIndex: matchLine.matchFileIndex,
								matchLineIndex: matchLine.matchLineIndex,
							}}
						/>
					);
				}
				return null;
			})}
		</ExpandableBlock>
	);
};
