"use server";

import { supportedCodemods, type SupportedCodemods } from "../../../codemods/codemods";
import fs from "fs";
import path from "path";

import allCodemodExports from "../../../codemods";

import { applyTransform } from "jscodeshift/src/testUtils";
import { getAllFilesAndFolders } from "../../../scanner/logic";
import type { LatestRunFile } from "../../../utils/latestFileRun";
import {
	addToOrReplaceFileInLatestRunFile,
	getPreviousStageOutput,
} from "../../../utils/latestFileRun";

/**
 * Run a set of codemods on a repo.
 * But **DOES NOT** update LatestRunFile with the result.
 */
export const runCodemods = async ({
	codemodsToRun = supportedCodemods,
	repoFolder,
}: {
	/**
	 * The codemods to run
	 * @default supportedCodemods
	 */
	codemodsToRun?: readonly SupportedCodemods[];
	repoFolder: string;
}) => {
	const includedFileExtensions = ["ts", "tsx", "js", "jsx"];

	const allFilesInRepo = await getAllFilesAndFolders({
		repoFolder,
		includedFileExtensions,
	});

	if (allFilesInRepo == null) {
		return;
	}

	const latestRunFile: LatestRunFile = {
		repoFolder,
		files: [],
		lastModified: Date.now(),
	};

	// Apply all codemods to all files
	// Promise.all is used to run all codemods in parallel and return the result in the same order
	// TODO: Consider replacing with allSettled to handle errors. API not 1-1 match
	let allCodemoddedFiles = await Promise.all(
		allFilesInRepo.map(async (filePath) => {
			return await runCodemod({
				filePath,
				codemodsToRun,
			});
		}),
	);

	// To avoid unnecessarily large JSON files, since
	// * formData fields are clipped at 2^20 characters. So the latestRunFile field can be clipped.
	// * Smaller JSON file = better app performance (read/write speed)
	// TODO: See if there is a better way to handle large JSON files
	allCodemoddedFiles = allCodemoddedFiles.filter(
		(f) => f.applicableCodemods.length > 0,
	) as LatestRunFile["files"];

	// Add all codemodded files to the latest run file
	for (const codemoddedFile of allCodemoddedFiles) {
		addToOrReplaceFileInLatestRunFile(latestRunFile, codemoddedFile.filePath, codemoddedFile);
	}

	return latestRunFile;
};

/**
 * Run a set of codemods on a single file.
 * But **DOES NOT** update LatestRunFile with the result.
 */
export const runCodemod = ({
	filePath,
	codemodsToRun = supportedCodemods,
}: {
	filePath: string;
	/**
	 * The codemods to run
	 * @default supportedCodemods
	 */
	codemodsToRun?: readonly SupportedCodemods[];
}): LatestRunFile["files"][number] => {
	// Trim to avoid unnecessary diff because of trailing whitespace/new lines
	const originalFileContents = fs.readFileSync(filePath, "utf8").trim();

	const codemoddedFile: LatestRunFile["files"][number] = {
		filePath,
		originalFileContents,
		state: "active",
		applicableCodemods: [],
		stages: [],
	};

	for (const codemod of codemodsToRun) {
		const previousCodeOutput =
			getPreviousStageOutput(codemoddedFile, codemoddedFile.stages.length) ?? originalFileContents;

		const currentCodeOutput = runCodemodTransform({
			codemod,
			filePath,
			codeInput: previousCodeOutput,
		});

		// Add stage only if the code output has been modified
		if (currentCodeOutput != null && currentCodeOutput !== previousCodeOutput) {
			codemoddedFile.stages.push({
				codemod: codemod,
				// If the codemod fails, use the previous code output
				output: currentCodeOutput,
			});
			codemoddedFile.applicableCodemods.push(codemod);
		}
	}

	return codemoddedFile;
};

/**
 * Run a single codemod on a single file
 * @returns The transformed code
 */
export const runCodemodTransform = ({
	codemod,
	filePath,
	codeInput,
}: {
	/**
	 * The codemod to run
	 */
	codemod: SupportedCodemods;
	/**
	 * File path of the file that will be passed to the codemod.
	 * - If you want to pass a *modified* version of this file to the codemod, also pass `codeInput`.
	 * - If you want to pass the *original* version of this file to the codemod, do not pass `codeInput`.
	 */
	filePath: string;
	/**
	 * Text/Code that will be passed to the codemod.
	 */
	codeInput?: string;
}) => {
	try {
		const codemodToRun = allCodemodExports[codemod];
		const codemodSource = codeInput ?? fs.readFileSync(filePath, "utf8");
		const applyTransformOutput = applyTransform(
			{
				default: codemodToRun,
				parser: "tsx",
			},
			{},
			{
				path: path.normalize(filePath),
				source: codemodSource,
			},
		);
		return applyTransformOutput;
	} catch (error) {
		console.error("ERROR:", error);
		return null;
	}
};

/**
 * Useful when the user changes their codemod filter and wants to re-run the codemods.
 * This function differs from runCodemod it retains essential parts of the previous run, instead of starting from scratch.
 * But **DOES NOT** update LatestRunFile with the result.
 *
 * @returns The new codemodded file
 */
export const reRunCodemod = async ({
	filePath,
	codemodsToRun,
	preExistingLatestRunFile,
}: {
	filePath: string;
	codemodsToRun: SupportedCodemods[];
	preExistingLatestRunFile: LatestRunFile;
}): Promise<LatestRunFile["files"][number] | null> => {
	const preExistingCodemoddedFileIndex =
		preExistingLatestRunFile?.files.findIndex((file) => file.filePath === filePath) ?? -1;
	const preExistingCodemoddedFile = preExistingLatestRunFile?.files[preExistingCodemoddedFileIndex];

	// Pre-existing codemodded file not found
	if (preExistingCodemoddedFile == null) {
		return null;
	}

	const newCodemoddedFile = await runCodemod({
		filePath,
		codemodsToRun,
	});

	// Pass preservable parts of the previous run to the new run
	newCodemoddedFile.applicableCodemods = preExistingCodemoddedFile.applicableCodemods;

	return newCodemoddedFile;
};

export const setFileStateToFinishedRunningCodemods = (file: LatestRunFile["files"][number]) => {
	if (file.state !== "active") {
		return;
	}

	const lastStageFileContents = getPreviousStageOutput(file, file.stages.length);
	fs.writeFileSync(file.filePath, lastStageFileContents);

	file.state = "finished-running";

	return file;
};

export const toggleIgnoreFile = (file: LatestRunFile["files"][number], shouldIgnore: boolean) => {
	if (file.state === "finished-running") {
		return;
	}

	file.state = shouldIgnore ? "ignored" : "active";

	return file;
};
