/*---------------------------------------------------------------------------------------------
 * Copyright (c) Bentley Systems, Incorporated. All rights reserved.
 * See LICENSE.md in the project root for license terms and full copyright notice.
 *--------------------------------------------------------------------------------------------*/
import { iTwinPackage, MatchFile } from "data/file_types";
import { exec } from "child_process";
import { MatchLine, MatchRule } from "data/file_types";
const fs = require("fs");
const dirTree = require("directory-tree");
import { DirectoryTree } from "directory-tree";
import path from "path";
const launch = require("launch-editor");

export const getAllFilesAndFoldersJSON = async (repoFolder: string) => {
	if (!repoFolder) {
		return null;
	}

	const allNonGitIgnoredAbsoluteFilePaths = await getAllFilesAndFolders({
		repoFolder,
	});
	if (!allNonGitIgnoredAbsoluteFilePaths) {
		return null;
	}

	const tree: DirectoryTree = dirTree(repoFolder, {
		nonGitIgnoredFiles: allNonGitIgnoredAbsoluteFilePaths,
	});
	return tree;
};

const readFile = (fileName: string): Promise<string> => {
	return new Promise((resolve, reject) => {
		fs.readFile(fileName, "utf8", (err: any, data: string) => {
			if (err) {
				console.error(err);
			}
			resolve(data);
		});
	});
};

const readFileGetLines = async (fileName: string) => {
	const fileContents = await readFile(fileName);
	return fileContents.split("\n");
};

export const getRulesFromFile = async (
	packagesToScan: iTwinPackage[] = ["itwinui-variables", "itwinui-css", "itwinui-react"],
): Promise<MatchRule[] | null> => {
	const rules: MatchRule[] = [];

	try {
		if (packagesToScan.includes("itwinui-variables")) {
			rules.push(...JSON.parse(await readFile(`${process.cwd()}/rules/rules_variables.json`)));
		}
		if (packagesToScan.includes("itwinui-css")) {
			rules.push(...JSON.parse(await readFile(`${process.cwd()}/rules/rules_css.json`)));
		}
		if (packagesToScan.includes("itwinui-react")) {
			rules.push(...JSON.parse(await readFile(`${process.cwd()}/rules/rules_react.json`)));
		}
	} catch (e) {
		return null;
	}

	return rules;
};

export const getMatchFile = async (
	fileName: string,
	rules: MatchRule[],
): Promise<
	| [
			{
				fileName: string;
				matchLines: MatchLine[];
			},
			number,
	  ]
	| null
> => {
	const matches: MatchLine[] = [];
	const fileLines = await readFileGetLines(fileName);

	if (rules == null) {
		return null;
	}

	for (const [lineIndex, line] of fileLines.entries()) {
		for (const rule of rules) {
			if (!rule.is_regex) {
				let index = line.indexOf(rule.replace);
				while (index >= 0) {
					index = line.indexOf(rule.replace, index + 1);

					matches.push({
						line_number: lineIndex + 1,
						columnNumber: index + 1,
						line: line,
						replace: rule.replace,
						replaceRule: rule.replace,
						with: rule.with,
						type: rule.type,
						description: rule.description,
						changed: false,
					});
				}
			} else {
				const lineMatches: IterableIterator<RegExpMatchArray> | null = line.matchAll(
					new RegExp(rule.replace, "g"),
				);
				if (lineMatches == null) {
					continue;
				}

				for (const lineMatch of lineMatches) {
					matches.push({
						line_number: lineIndex + 1,
						columnNumber: (lineMatch.index ?? 0) + 1,
						line: line,
						replace: lineMatch[0],
						replaceRule: rule.replace,
						with: rule.with,
						type: rule.type,
						description: rule.description,
						changed: false,
					});
				}
			}
		}
	}

	const matchObject = {
		fileName: fileName,
		matchLines: matches,
	};
	return [matchObject, fileLines.length];
};

const implementLineMatchChange = (lines: string[], lineMatch: MatchLine, shouldUndo: boolean) => {
	if (
		(!shouldUndo && lines[lineMatch.line_number - 1] !== lineMatch.line) ||
		(shouldUndo &&
			lines[lineMatch.line_number - 1] !==
				lineMatch.line.replace(lineMatch.replace, lineMatch.with))
	) {
		return {
			success: false,
			lines: lines,
			changed: shouldUndo,
		};
	}
	if (lineMatch.type !== "autoswap") {
		return {
			success: false,
			lines: lines,
			changed: shouldUndo,
		};
	}

	lines[lineMatch.line_number - 1] = lines[lineMatch.line_number - 1].replace(
		!shouldUndo ? lineMatch.replace : lineMatch.with,
		!shouldUndo ? lineMatch.with : lineMatch.replace,
	);

	return {
		success: true,
		lines: lines,
		changed: !shouldUndo,
	};
};

export const implementLineMatchChanges = async ({
	fileName,
	lineMatches,
	shouldUndo,
}: {
	fileName: string;
	lineMatches: MatchLine[];
	shouldUndo: boolean;
}) => {
	return new Promise(async (resolve, reject) => {
		const lines = await readFileGetLines(fileName);
		const successes = [];
		const changedLines = [];

		for (let lineMatch of lineMatches) {
			const result = implementLineMatchChange(lines, lineMatch, shouldUndo);
			successes.push(result.success);
			changedLines.push(result.changed);
		}

		fs.writeFile(fileName, lines.join("\n"), (err: any, data: any) => {
			if (err) {
				console.error(err);
			}
		});

		resolve({
			successes,
			changedLines,
		});
	});
};

export const getMatchFiles = async (
	fileNames: string[],
	packagesToScan: iTwinPackage[],
): Promise<MatchFile[] | null> => {
	const fileMatches = [];

	let numFilesScanned = 0;
	let numLinesScanned = 0;

	const rules: MatchRule[] | null = await getRulesFromFile(packagesToScan);
	if (rules == null) {
		return null;
	}

	for (const fileName of fileNames) {
		const stats = fs.statSync(fileName);
		if (!stats.isFile()) {
			continue;
		}

		const fileExtension = fileName.split(".").pop();

		// Skip image type files
		if (["png", "jpg", "gif", "PNG", undefined].includes(fileExtension)) {
			continue;
		}

		const output = await getMatchFile(fileName, rules);
		if (output == null) {
			return null;
		}

		const [fileMatch, numLines] = output;
		numFilesScanned++;
		numLinesScanned += numLines;

		// Push only if at least one match is found in the file
		if (fileMatch.matchLines.length > 0) {
			fileMatches.push(fileMatch);
		}
	}

	return fileMatches;
};

export const isInGitWorkingTree = (repoPath: string): Promise<boolean> => {
	return new Promise<boolean>((resolve, reject) => {
		const command = `cd "${repoPath}" && git rev-parse --is-inside-work-tree`;
		exec(command, (error, stdout, stderr) => {
			if (error) {
				console.error("error: ", error);
				resolve(false);
				return;
			}

			const isInsideWorkingTree = stdout.trim() === "true";
			resolve(isInsideWorkingTree);
		});
	});
};

export const getGitTopLevelPath = (repoPath: string): Promise<string | null> => {
	return new Promise<string | null>((resolve, reject) => {
		const command = `cd "${repoPath}" && git rev-parse --show-toplevel`;
		exec(command, (error, stdout, stderr) => {
			if (error) {
				console.error("error: ", error);
				resolve(null);
				return;
			}

			resolve(stdout.trim());
		});
	});
};

export const getAllFilesAndFolders = async ({
	repoFolder,
	includedFileExtensions,
}: {
	repoFolder: string;
	/**
	 * File extensions to include when scanning the repo.
	 *
	 * To include all extensions, pass undefined (or don't pass anything).
	 *
	 * @default
	 * undefined
	 *
	 * @example
	 * <caption>Include only `.ts` and `.tsx` files</caption>
	 * ["ts", "tsx"]
	 */
	includedFileExtensions?: string[];
}) => {
	if (await isInGitWorkingTree(repoFolder)) {
		return getAllFilesAndFoldersFromGitRepo(repoFolder, includedFileExtensions);
	} else {
		return getAllFilesAndFoldersFromNonGitRepo(repoFolder, includedFileExtensions);
	}
};

const getAllFilesAndFoldersFromGitRepo = (
	repoFolder: string,
	includedFileExtensions?: string[],
) => {
	return new Promise<string[] | null>((resolve, reject) => {
		const command = `cd "${repoFolder}" && git ls-files --full-name`;
		exec(command, async (error, stdout, stderr) => {
			if (error) {
				console.error("error: ", error);
				resolve(null);
				return;
			}

			const gitTopLevelPath = await getGitTopLevelPath(repoFolder);
			if (gitTopLevelPath == null) {
				resolve(null);
				return;
			}

			// Prepending gitTopLevelPath since `git ls-files --full-name` returns paths relative to the git top level path
			// But we want absolute paths
			const allNonGitIgnoredAbsoluteFilePaths = stdout
				.trim()
				.split("\n")
				.map((filePath) => path.join(gitTopLevelPath, filePath))
				.filter((filePath) => {
					const fileExtension = path.extname(filePath).substring(1); // To ignore the . in the file extension
					return includedFileExtensions?.includes(fileExtension) ?? true;
				});

			resolve(allNonGitIgnoredAbsoluteFilePaths);
		});
	});
};

const getAllFilesAndFoldersFromNonGitRepo = (
	repoFolder: string,
	includedFileExtensions?: string[],
) => {
	return new Promise<string[] | null>((resolve, reject) => {
		let command = `cd "${repoFolder}" && find . `;

		if (includedFileExtensions != null) {
			// Include all files with file extensions  in `includedFileExtensions`
			command += ` \\( `;
			includedFileExtensions.forEach((pattern, index) => {
				if (index !== 0) {
					command += `-o`;
				}
				command += ` -name "*.${pattern}" `;
			});
			command += `\\)`;
		}

		// Print only all the files (`-type f`)
		command += ` -type f -print`;

		exec(command, (error, stdout, stderr) => {
			if (error) {
				console.error("error: ", error);
				resolve(null);
				return;
			}

			// Prepending repoFolder since git ls-files returns repo relative paths
			const allNonGitIgnoredAbsoluteFilePaths = stdout
				.trim()
				.split("\n")
				.map((filePath) => path.join(repoFolder, filePath));

			resolve(allNonGitIgnoredAbsoluteFilePaths);
		});
	});
};

export const openFile = async (fileName: string, lineNumber: number, columnNumber: number) => {
	try {
		launch(`${fileName}:${lineNumber}:${columnNumber}`);
		return true;
	} catch {
		return false;
	}
};
