import { FolderInfo } from "data/file_types";
import directoryTree from "directory-tree";
import { iTwinPackage, MatchFile, MatchRule } from "data/file_types";
import { displayToast } from "./tasks";
import { SupportedCodemods } from "codemods/codemods";

export const getMatchesInRepo = async (
	gettingMatches: boolean,
	repoFolder: string,
	packagesToScan: iTwinPackage[],
): Promise<MatchFile[] | null> => {
	return getMatches({ gettingMatches, repoFolder, packagesToScan });
};

/**
 *
 * @param gettingMatches
 * @param fileName
 * @param packagesToScan
 * @returns `[MatchFile]` when there exists at least one matchLine in file; `[]` when no matchLines found in file; `null` when error finding matches
 */
export const getMatchesInFile = async (
	gettingMatches: boolean,
	fileName: string,
	packagesToScan: iTwinPackage[],
): Promise<MatchFile[] | null> => {
	const matches = await getMatches({
		gettingMatches,
		fileName,
		packagesToScan,
	});
	if (matches != null) {
		if (matches.length === 1) {
			return matches;
		} else {
			return [];
		}
	}
	return null;
};

const getPackagesToScanQueryParameters = (packagesToScan: iTwinPackage[]) => {
	return packagesToScan.map((p) => `packagesToScan=${p}`);
};

export const getFolderInfo = async (repoFolder: string): Promise<FolderInfo | null> => {
	const responseFolderInfo = await fetch(`/api/folder_info?repoFolder=${repoFolder}`, {
		method: "GET",
		headers: { "Content-Type": "application/json" },
	});

	let folderInfo: FolderInfo | null = null;
	if (responseFolderInfo.ok) {
		folderInfo = await responseFolderInfo.json();
	}

	return folderInfo;
};

/**
 * Helper method: Get matches from scanner.
 *
 * `assert(fileName != null || folderName != null)`
 * @param gettingMatches
 * @param fileName
 * @param repoFolder
 * @returns `Promise<MatchFile[] | null>`
 */
const getMatches = async ({
	gettingMatches,
	fileName,
	repoFolder,
	packagesToScan,
}: {
	gettingMatches: boolean;
	fileName?: string;
	repoFolder?: string;
	packagesToScan: iTwinPackage[];
}): Promise<MatchFile[] | null> => {
	if (gettingMatches) {
		return null;
	}

	const queryParam = fileName != null ? `fileName=${fileName}` : `repoFolder=${repoFolder}`;
	const packagesToScanQueryParams = getPackagesToScanQueryParameters(packagesToScan);

	const response = await fetch(`/api/matches?${queryParam}&${packagesToScanQueryParams}`, {
		method: "GET",
		headers: { "Content-Type": "application/json" },
	});

	let matches = null;
	if (response.ok) {
		matches = await response.json();
	} else {
		displayToast("negative", "Error finding matches");
	}

	return matches;
};

export const getFolders = async (
	repoFolder: string,
): Promise<directoryTree.DirectoryTree | null> => {
	const response = await fetch(`/api/files?repoFolder=${repoFolder}`, {
		method: "GET",
		headers: { "Content-Type": "application/json" },
	});

	let folders = null;
	if (response.ok) {
		folders = await response.json();
	}

	return folders;
};

export const getRules = async (packagesToScan: iTwinPackage[]): Promise<MatchRule[] | null> => {
	const packagesToScanQueryParams = getPackagesToScanQueryParameters(packagesToScan);

	const response = await fetch(`/api/rules?${packagesToScanQueryParams.join("&")}`, {
		method: "GET",
		headers: { "Content-Type": "application/json" },
	});

	let rules = null;
	if (response.ok) {
		rules = await response.json();
	} else {
		displayToast(
			"negative",
			"Error scanning the provided folder. Try ensuring the folder is valid.",
		);
	}

	return rules;
};
