import j, { ASTPath, ImportSpecifier, JSXElement } from "jscodeshift";
import {
	filterOutPropFromComponent,
	getITwinUIReactImportDeclarations,
	getImportAliasToRealComponentMapping,
	getImportSpecifierName,
} from "./codemods";
import { describe, test, expect } from "vitest";

describe("util functions", () => {
	test("getITwinUIReactImportDeclarations", () => {
		const code = `
import { Button, Alert } from "foobar";
import { Carousel, Table } from "@itwin/itwinui-react";
`.trim();
		const root = j(code);

		const expectedImportSpecifiers = ["Carousel", "Table"];
		const actualImportSpecifiers = getITwinUIReactImportDeclarations(j, root)
			.paths()
			.flatMap((importDeclaration) => importDeclaration.node.specifiers);

		expect(expectedImportSpecifiers.length).toBe(actualImportSpecifiers.length);
		for (let i = 0; i < expectedImportSpecifiers.length; i++) {
			const actualSpecifierNames = getImportSpecifierName(
				actualImportSpecifiers[i] as ImportSpecifier,
			);
			expect(actualSpecifierNames.importedName).toBe(expectedImportSpecifiers[i]);
		}
	});

	test("getImportAliasToRealComponentMapping", () => {
		const code = `
import { Button, Alert as MyAlert } from "@itwin/itwinui-react";
import { Header as MyHeader, Tree as MyTree } from "@itwin/itwinui-react";
import { Tree } from "foobar";
`.trim();
		const root = j(code);

		// When `components` are not passed,
		// we should get all iTwinUI components
		expect(getImportAliasToRealComponentMapping(j, root)).toStrictEqual({
			Button: "Button",
			MyAlert: "Alert",
			MyHeader: "Header",
			MyTree: "Tree",
		});

		// When `components` are passed,
		// we should only get key-value pairs where the component's real name matches the components array
		expect(getImportAliasToRealComponentMapping(j, root, ["Button", "Tree"])).toStrictEqual({
			Button: "Button",
			MyTree: "Tree",
		});
	});

	test("filterOutPropFromComponent", () => {
		const inputCode = `<LabeledInput label="My label" setFocus={false} />`.trim();
		const expectedOutputCode = `<LabeledInput label="My label" />`.trim();

		const root = j(inputCode);

		const labeledInput = root.find(JSXElement).get(0) as ASTPath<JSXElement>;
		labeledInput.node.openingElement.attributes = filterOutPropFromComponent(
			labeledInput,
			"setFocus",
		);

		const actualOutputCode = root.toSource();

		expect(actualOutputCode).toEqual(expectedOutputCode);
	});
});
