import {
	ASTPath,
	FileInfo,
	JSXAttribute,
	JSXElement,
	JSXEmptyExpression,
	JSXExpressionContainer,
	Node,
	ObjectExpression,
	ObjectProperty,
	SourceLocation,
	API,
	Collection,
	ImportDeclaration,
	BlockStatement,
	JSXOpeningElement,
} from "jscodeshift";
import fs from "fs";
import { Suggestion } from "data/codemod_types";
export const suggestionsFilePath = "logs/v3/suggestions.json" as const;
import { isRunningFromUnitTest } from "../utils/tests";
import { SupportedCodemods } from "codemods/codemods";

export const createLogsFolderIfNotExists = (): void => {
	if (!fs.existsSync("logs")) {
		fs.mkdirSync("logs");
	}
	if (!fs.existsSync("logs/v3")) {
		fs.mkdirSync("logs/v3");
	}
};

export const createEmptySuggestionsFile = () => {
	createLogsFolderIfNotExists();

	fs.writeFileSync(suggestionsFilePath, "[]");
};

/**
 * Adds a suggestion to the suggestionsFilePath file. This is useful for one-time alerts
 * that need to persist after the codemod is run since re-running the codemods will not
 * re-produce this alert.
 */
export const addSuggestionToSuggestionsFile = ({
	fileInfo,
	loc,
	message,
	codemod,
}: {
	fileInfo: FileInfo;
	loc: SourceLocation | null | undefined;
	message: string;
	codemod: SupportedCodemods;
}): void => {
	// Skip if in unit test
	if (isRunningFromUnitTest) {
		return;
	}

	const allLines = fileInfo.source.split("\n");

	let lineContent = undefined;
	let linesBefore = undefined;
	let linesAfter = undefined;

	if (loc != null) {
		// Start line only
		lineContent = allLines[loc.start.line - 1];

		// Max 5 lines before and after
		linesBefore = allLines.slice(0, loc.start.line - 1).slice(-5);
		linesAfter = allLines.slice(loc.end.line).slice(0, 5);
	}

	// Create a unique id that is unique all the time
	const uniqueId = Math.random().toString(36).substring(2);

	// Add a suggestion to the suggestionsFilePath file
	const suggestions = fs.readFileSync(suggestionsFilePath, "utf-8");
	const newSuggestion: Suggestion = {
		id: uniqueId,
		dismissed: false,
		timestamp: new Date().toISOString(),
		type: "suggestion",
		file: fileInfo.path,
		codemod: codemod,
		message: {
			content: message,
			start: {
				line: loc?.start.line ?? 0,
				column: (loc?.start.column ?? 0) + 1,
			},
			lineContent,
			linesBefore,
			linesAfter,
		},
	};
	const newSuggestions = [...JSON.parse(suggestions), newSuggestion];
	fs.writeFileSync("logs/v3/suggestions.json", JSON.stringify(newSuggestions, null, 2));
};

/*
 * @example
 * ```ts
 * <Checkbox
 *  label = 'Checkbox label'
 *  className={"my-class"}
 *  style={{ color: "red" }}
 * />
 * ```
 *
 * ```ts
 * const className = getPropFromComponent(wrapperProps, "className");
 * // className = "my-class"
 *
 * const style = getPropFromComponent(wrapperProps, "style");
 * // style = { color: "red" }
 * ```
 */
export const getPropFromComponent = (
	path: ASTPath<JSXElement> | ASTPath<JSXOpeningElement>,
	propName: string,
): JSXAttribute | undefined => {
	const openingElement = path.node.type === "JSXElement" ? path.node.openingElement : path.node;

	const propValue = (
		openingElement.attributes?.filter((attr) => attr?.type === "JSXAttribute") as JSXAttribute[]
	)?.find((attr) => {
		if (attr.name.type === "JSXIdentifier" && attr.name.name === propName) {
			return attr;
		}
	});

	return propValue;
};

/**
 * Returns a new list of attributes with the `propName` removed.
 */
export const filterOutPropFromComponent = (
	path: ASTPath<JSXElement> | ASTPath<JSXOpeningElement>,
	propName: string,
): JSXAttribute[] => {
	const openingElement = path.node.type === "JSXElement" ? path.node.openingElement : path.node;

	const newAttributes = (
		openingElement.attributes?.filter((attr) => attr.type === "JSXAttribute") as JSXAttribute[]
	).filter((attr) => !(attr.name.type === "JSXIdentifier" && attr.name.name === propName));

	return newAttributes;
};

/**
 * Renames a component to `newName`.
 *
 * @returns hasChanged
 */
export const renameComponent = (path: ASTPath<JSXElement>, newName: string): boolean => {
	let hasChanged = false;

	if (path.node.openingElement.name.type !== "JSXMemberExpression") {
		path.node.openingElement.name.name = newName;
		hasChanged = true;
	}
	if (
		path.node.closingElement != null &&
		path.node.closingElement.name.type !== "JSXMemberExpression"
	) {
		path.node.closingElement.name.name = newName;
		hasChanged = true;
	}

	return hasChanged;
};

/**
 * Returns the ObjectProperty from the ObjectExpression if the key matches
 *
 * @example
 * ```ts
 * const wrapperProps = { className: "my-class", style: { color: "red" } };
 * const className = getObjectPropertyFromObjectExpression(wrapperProps, "className");
 * // className = "my-class"
 * const style = getObjectPropertyFromObjectExpression(wrapperProps, "style");
 * // style = { color: "red" }
 * ```
 */
export const getObjectPropertyFromObjectExpression = (
	objectExpression: ObjectExpression,
	key: string,
): ObjectProperty | undefined => {
	let objectProperty: ObjectProperty | undefined;

	objectExpression.properties.forEach((obj) => {
		if (obj.type === "ObjectProperty" && obj.key.type === "Identifier") {
			if (obj.key.name === key) {
				objectProperty = obj;
			}
		}
	});

	return objectProperty;
};

export const filterOutObjectPropertyFromObjectExpression = (
	objectExpression: ObjectExpression,
	key: string,
): ObjectProperty[] => {
	const newProperties = (
		objectExpression.properties.filter((prop) => prop.type === "ObjectProperty") as ObjectProperty[]
	).filter((prop) => {
		if (prop.type === "ObjectProperty" && prop.key.type === "Identifier") {
			if (prop.key.name === key) {
				return false;
			}
		}
		return true;
	});

	return newProperties;
};

/**
 * Returns the simplest logical form of the value of the JSXAttribute, whenever possible.
 * E.g. if the value is a JSXExpressionContainer, then it returns the expression.
 *
 * If nothing logical simplification can be done, it just returns jsxAttribute.value.
 *
 * @example
 * <Component className="my-class" />
 * // return = "my-class"
 * <Component className={"my-class"} />
 * // return = "my-class" (not {"my-class"})
 * <Component className={myClass} />
 * // return = myClass (not {myClass})
 * <Component className={myClass + "my-class-2"} />
 * // return = myClass + "my-class-2" (not {myClass + "my-class-2"})
 */
export const getSimplifiedJSXAttributeValue = (jsxAttribute: JSXAttribute | null | undefined) => {
	if (jsxAttribute == null) {
		return null;
	}

	type ReturnType =
		| JSXAttribute["value"]
		// If the expression is a JSXEmptyExpression, then we should not return it.
		// This is because it causes errors when adding the return value to a j.ObjectProperty
		| Exclude<JSXExpressionContainer["expression"], JSXEmptyExpression>;

	let value: ReturnType = jsxAttribute.value;

	if (jsxAttribute.value != null) {
		if (jsxAttribute.value.type === "StringLiteral") {
			value = jsxAttribute.value;
		} else if (jsxAttribute.value.type === "JSXExpressionContainer") {
			// If the expression is a JSXEmptyExpression, then we should not return it.
			if (jsxAttribute.value.expression.type !== "JSXEmptyExpression") {
				value = jsxAttribute.value.expression;
			}
		}
	}

	return value;
};

/**
 * Returns a shortened version of the ASTPath.node.
 *
 * Useful for debugging codemods.
 */
export const getShortenedDebugASTPath = (node: Node) => {
	const debugJson = node;

	// Remove all the "loc" keys from the debugJson and its sub-objects.
	// This is done because the debugJson is too big and it makes the test output unreadable.
	const removeLoc = (obj: any) => {
		if (obj == null) {
			return;
		}

		if (obj.loc != null) {
			delete obj.loc;
		}
		Object.keys(obj).forEach((key) => {
			if (typeof obj[key] === "object") {
				removeLoc(obj[key]);
			}
		});
	};
	removeLoc(debugJson);

	return debugJson;
};

/**
 * @example
 * import { HorizontalTabs as MyHorizontalTabs } from "@itwin/itwinui-react";
 * getImportSpecifierName() = {
 *   "name": "MyHorizontalTabs",
 *   "importedName": "HorizontalTabs",
 * }
 *
 * @example
 * import { VerticalTabs } from "@itwin/itwinui-react";
 * getImportSpecifierName() = {
 *  "name": "VerticalTabs",
 *  "importedName": "VerticalTabs",
 * }
 */
export const getImportSpecifierName = (
	importSpecifier: NonNullable<ASTPath<ImportDeclaration>["node"]["specifiers"]>[number],
): {
	/**
	 * @example
	 * import { HorizontalTabs as MyHorizontalTabs } from "@itwin/itwinui-react";
	 * // name = "MyHorizontalTabs"
	 * // importedName = "HorizontalTabs"
	 *
	 * @example
	 * import { VerticalTabs } from "@itwin/itwinui-react";
	 * // name = "VerticalTabs"
	 * // importedName = "VerticalTabs"
	 */
	name: string | null | undefined;
	/**
	 * @example
	 * import { HorizontalTabs as MyHorizontalTabs } from "@itwin/itwinui-react";
	 * // name = "MyHorizontalTabs"
	 * // importedName = "HorizontalTabs"
	 *
	 * @example
	 * import { VerticalTabs } from "@itwin/itwinui-react";
	 * // name = "VerticalTabs"
	 * // importedName = "VerticalTabs"
	 */
	importedName: string | null | undefined;
} => {
	const name = importSpecifier.local?.name;
	const importedName = (importSpecifier as any)?.imported?.name;

	return {
		name,
		importedName,
	};
};

/**
 * Returns all import declarations from `@itwin/itwinui-react`
 */
export const getITwinUIReactImportDeclarations = (
	j: API["jscodeshift"],
	/**
	 * Example value: `root`
	 */
	collection: Collection,
) => {
	return collection.find(j.ImportDeclaration, {
		source: {
			value: "@itwin/itwinui-react",
		},
	});
};

/**
 * Removes the import specifiers from all import declarations.
 * When import aliases are used, `specifiersToRemove` will only be compared against the real component names, and not the aliases.
 *
 * @returns `[numberOfImportSpecifiersRemoved, hasChanged]`
 *
 * @example
 * removeImportSpecifiers(j, root, ["HorizontalTabs", "VerticalTabs", "ShouldNotRemove"]);
 *
 * // Before
 * import { Button, HorizontalTabs } from "@itwin/itwinui-react";
 * import { VerticalTabs as MyVerticalTabs } from "@itwin/itwinui-react";
 * import { Text as ShouldNotRemove } from "@itwin/itwinui-react";
 *
 * // After
 * import { Button } from "@itwin/itwinui-react";
 * import { Text as ShouldNotRemove } from "@itwin/itwinui-react";
 */
export const removeImportSpecifiers = (
	j: API["jscodeshift"],
	root: ReturnType<typeof j>,
	specifiersToRemove: string[],
): [number, boolean] => {
	let numberOfImportSpecifiersRemoved = 0;
	let hasChanged = false;

	const iTwinUiImportDeclarations = getITwinUIReactImportDeclarations(j, root);
	iTwinUiImportDeclarations.forEach((importDeclaration) => {
		const oldImportSpecifiers =
			importDeclaration.node.specifiers != null
				? [...importDeclaration.node.specifiers]
				: undefined;
		const newImportSpecifiers = importDeclaration.node.specifiers?.filter((specifier) => {
			const specifierImportedName = getImportSpecifierName(specifier).importedName;

			const isSpecifierImportedNameToRemove =
				specifierImportedName != null ? specifiersToRemove.includes(specifierImportedName) : false;

			return !isSpecifierImportedNameToRemove;
		});

		if (oldImportSpecifiers == null || newImportSpecifiers == null) {
			return;
		}

		// Update hasChanged
		hasChanged = hasChanged || oldImportSpecifiers.length !== newImportSpecifiers.length;

		// Update importDeclaration.node.specifiers
		importDeclaration.node.specifiers = newImportSpecifiers;

		// Update numberOfImportSpecifiersRemoved
		if (oldImportSpecifiers != null && newImportSpecifiers != null) {
			numberOfImportSpecifiersRemoved += oldImportSpecifiers.length - newImportSpecifiers.length;
		}

		// If there are no specifiers left in the import declaration, remove the import declaration
		if (importDeclaration.node.specifiers?.length === 0) {
			importDeclaration.prune();
			hasChanged = true;
		}
	});

	return [numberOfImportSpecifiersRemoved, hasChanged];
};

/**
 * Adds the import specifiers to the first import declaration, if the import specifiers do not already exist.
 * When import aliases are used, `specifiersToAdd` will only be compared against the alias names and not the real component names.
 *
 * @returns `[numberOfImportSpecifiersAdded, hasChanged]`
 *
 * @example
 * addImportSpecifiers(j, root, ["Tabs", "Button"]);
 *
 * // Before
 * import { Button } from "@itwin/itwinui-react";
 * import { Tabs as MyPreexistingTabs } from "@itwin/itwinui-react";
 *
 * // After
 * import { Button, Tabs } from "@itwin/itwinui-react";
 * import { Tabs as MyPreexistingTabs } from "@itwin/itwinui-react";
 */
export const addImportSpecifiers = (
	j: API["jscodeshift"],
	root: ReturnType<typeof j>,
	specifiersToAdd: string[],
): [number, boolean] => {
	let numberOfImportSpecifiersAdded = 0;
	let hasChanged = false;

	try {
		let iTwinUiImportDeclarations = getITwinUIReactImportDeclarations(j, root);
		const existingImportSpecifiers = new Set<string>();

		// Find all existing import specifiers, using the alias names
		iTwinUiImportDeclarations.forEach((importDeclaration) => {
			const importSpecifiers = importDeclaration.node.specifiers;
			importSpecifiers?.forEach((specifier) => {
				const specifierName = getImportSpecifierName(specifier).name;

				if (specifierName != null) {
					existingImportSpecifiers.add(specifierName);
				}
			});
		});

		// Find specifiersRemainingToAdd
		const specifiersRemainingToAdd = specifiersToAdd.filter(
			(specifierToAdd) => !existingImportSpecifiers.has(specifierToAdd),
		);

		// If specifiersRemainingToAdd is empty, return
		if (specifiersRemainingToAdd.length === 0) {
			throw null;
		}

		// If no import declarations from @itwin/itwinui-react, add an empty one
		if (iTwinUiImportDeclarations.length === 0) {
			root
				.find(j.ImportDeclaration)
				.at(0)
				.insertBefore(j.importDeclaration([], j.literal("@itwin/itwinui-react")));
			hasChanged = true;
		}
		iTwinUiImportDeclarations = getITwinUIReactImportDeclarations(j, root);

		// Add specifiersRemainingToAdd to the first import declaration
		iTwinUiImportDeclarations
			.nodes()[0]
			.specifiers?.push(
				...specifiersRemainingToAdd.map((specifierToAdd) =>
					j.importSpecifier(j.identifier(specifierToAdd)),
				),
			);
		numberOfImportSpecifiersAdded = specifiersRemainingToAdd.length;
		hasChanged = true;
	} finally {
		return [numberOfImportSpecifiersAdded, hasChanged];
	}
};

/**
 * Returns a mapping of the aliases of the components to their actual names.
 *
 * @example
 * import { Button, Alert as MyAlert } from "@itwin/itwinui-react";
 * import { Header as MyHeader, Tree as MyTree } from "@itwin/itwinui-react";
 *
 * getAliasMapping(j, root, ["Button", "Alert", "Tree"]):
 * // Returns:
 * // {
 * //   Button: "Button",
 * //   MyAlert: "Alert",
 * //   MyTree: "Tree",
 * // }
 */
export const getImportAliasToRealComponentMapping = (
	j: API["jscodeshift"],
	/**
	 * Example value: `root`
	 */
	collection: Collection,
	/**
	 * If provided, only the components in this array will be considered.
	 * If not provided, all components will be considered.
	 */
	components?: string[],
) => {
	const aliasMapping: Record<string, string> = {};

	getITwinUIReactImportDeclarations(j, collection).forEach((importDeclaration) => {
		importDeclaration.node.specifiers?.forEach((specifier) => {
			// E.g. #1: import { Body as BodyOld } from "@itwin/itwinui-react";
			// E.g. #2: import { Body } from "@itwin/itwinui-react";
			const specifierNames = getImportSpecifierName(specifier);
			const specifierName = specifierNames.name; // #1: BodyOld | #2: Body
			const specifierImportedName = specifierNames.importedName; // #1: Body | #2: Body

			// If the component is not in the components array, skip it
			if (
				components != null &&
				specifierImportedName != null &&
				!components.includes(specifierImportedName)
			) {
				return;
			}

			// Else, add it to the aliasMapping
			if (specifierName != null && specifierImportedName != null) {
				aliasMapping[specifierName] = specifierImportedName;
			}
		});
	});

	return aliasMapping;
};

/**
 * Renames the oldProp (deprecated) to newProp (replacement).
 *
 * Cases:
 * - if oldProp is present and newProp is not present: renames oldProp to newProp
 * - if both oldProp and newProp are present: gives a warning that both the deprecated prop and its replacement are being used
 *
 * @returns hasChanged
 */
export const renameProp = ({
	fileInfo,
	path,
	oldPropName,
	newPropName,
	seeLink,
	codemod,
}: {
	fileInfo: FileInfo;
	path: ASTPath<JSXElement>;
	oldPropName: string;
	newPropName: string;
	/**
	 * Optional link to see more info about the prop rename
	 */
	seeLink?: string;
	codemod: SupportedCodemods;
}): boolean => {
	let hasChanged = false;

	const oldPropValue = getPropFromComponent(path, oldPropName);
	if (oldPropValue != null) {
		const newPropValue = getPropFromComponent(path, newPropName);
		if (newPropValue != null) {
			// Give a warning that both oldProp and newProp are being used
			let message = `Both \`${oldPropName}\` and \`${newPropName}\` are being used. Please remove the \`${oldPropName}\` prop and only use the \`${newPropName}\` prop instead.`;
			if (seeLink != null) {
				message += ` (See [link](${seeLink}))`;
			}

			addSuggestionToSuggestionsFile({
				fileInfo,
				loc: path.node.loc,
				message,
				codemod,
			});
		} else {
			// Rename oldProp to newProp
			oldPropValue.name.name = newPropName;
			hasChanged = true;
		}
	}

	return hasChanged;
};

/*
 * Returns the ASTPath of the n-th ancestor BlockStatement, if it exists.
 * Else, returns null.
 */
export const getAncestorBlockStatement = (
	path: ASTPath<any>,
	/**
	 * Since this function returns the n-th ancestor BlockStatement, this parameter
	 * specifies the value of n.
	 *
	 * Set it to "upper-most" to return the upper-most ancestor BlockStatement.
	 *
	 * @default 1
	 */
	ancestorOrdinalNumber: number | "upper-most" = 1,
): ASTPath<BlockStatement> | null => {
	let currentNode = path;
	let ancestorBlockStatements = [];

	// Invalid ancestorOrdinalNumber
	if (typeof ancestorOrdinalNumber === "number" && ancestorOrdinalNumber < 1) {
		return null;
	}

	// Find all BlockStatements in the ancestor chain
	while (currentNode.parent != null) {
		currentNode = currentNode.parent;

		if (currentNode.node.type === "BlockStatement") {
			ancestorBlockStatements.push(currentNode as ASTPath<BlockStatement>);

			// If BlockStatement at ancestorOrdinalNumber is found, return it
			if (
				typeof ancestorOrdinalNumber === "number" &&
				ancestorOrdinalNumber === ancestorBlockStatements.length
			) {
				return currentNode as ASTPath<BlockStatement>;
			}
		}
	}

	// If ancestorOrdinalNumber is "upper-most", return the upper-most ancestor BlockStatement
	if (ancestorOrdinalNumber === "upper-most") {
		return ancestorBlockStatements[ancestorBlockStatements.length - 1] ?? null;
	}

	// Else, return the n-th ancestor BlockStatement
	return ancestorBlockStatements[ancestorOrdinalNumber - 1] ?? null;
};

/**
 * Returns true if the path is inside a class component.
 */
export const isInClassComponent = (path: ASTPath<any>) => {
	let currentNode = path;

	while (currentNode.parent != null) {
		currentNode = currentNode.parent;

		if (currentNode.node.type === "ClassDeclaration") {
			return true;
		}
	}

	return false;
};

/**
 * Moves the className and the style props from the component to the wrapperProps prop.
 * - If wrapperProps already exists, then we do nothing, since we assume that the user already knows about wrapperProps
 * and wants to pass the className and style to the inner main element and not the wrapper.
 * - If wrapperProps does not exist, then we create a new wrapperProps with the className and style.
 *
 * This method is useful for components like:
 * - `Checkbox`: [#1371](https://github.com/iTwin/iTwinUI/pull/1371)
 * - `RadioTile`: [#1409](https://github.com/iTwin/iTwinUI/pull/1409)
 * - `SideNavigation`: [#1458](https://github.com/iTwin/iTwinUI/pull/1458)
 *
 * @returns hasChanged
 */
export const moveClassNameAndStylePropsToWrapperProps = ({
	j,
	root,
	componentName: componentNameProp,
	excludeCondition,
}: {
	j: API["jscodeshift"];
	root: ReturnType<typeof j>;
	/**
	 * E.g. `RadioTile`, `Checkbox`, `SideNavigation`
	 */
	componentName: string;
	excludeCondition?: (path: ASTPath<JSXElement>) => boolean;
}): boolean => {
	const importAliasToRealComponentMapping = getImportAliasToRealComponentMapping(j, root, [
		componentNameProp,
	]);
	let hasChanged = false;

	// Run for all aliases of componentName
	Object.keys(importAliasToRealComponentMapping).forEach((componentName) => {
		const components = root.findJSXElements(componentName);

		components.forEach((componentPath) => {
			// If excludeCondition is provided, skip the component if the condition is true
			if (excludeCondition != null && excludeCondition(componentPath)) {
				return;
			}

			const component = componentPath.node;

			if (component.openingElement.attributes == null) {
				return;
			}

			const classNameProp = component.openingElement.attributes.find(
				(attr) => attr.type === "JSXAttribute" && attr.name.name === "className",
			) as JSXAttribute | undefined;
			const styleProp = component.openingElement.attributes.find(
				(attr) => attr.type === "JSXAttribute" && attr.name.name === "style",
			) as JSXAttribute | undefined;

			if (!classNameProp && !styleProp) {
				return;
			}

			const wrapperPropsProp = component.openingElement.attributes.find(
				(attr) => attr.type === "JSXAttribute" && attr.name.name === "wrapperProps",
			) as JSXAttribute | undefined;

			if (wrapperPropsProp != null) {
				// We should do nothing, since looks like the user already knows about wrapperProps.
				// i.e. The user really wants to pass the className to the inner main element and not the wrapper.
			} else {
				const classNamePropValue = getSimplifiedJSXAttributeValue(classNameProp);
				const stylePropValue = getSimplifiedJSXAttributeValue(styleProp);

				// Add wrapperProps (with className/style) to component
				component.openingElement.attributes.push(
					j.jsxAttribute(
						j.jsxIdentifier("wrapperProps"),
						j.jsxExpressionContainer(
							j.objectExpression(
								[
									classNamePropValue != null &&
										j.objectProperty(j.identifier("className"), classNamePropValue),
									stylePropValue != null && j.objectProperty(j.identifier("style"), stylePropValue),
								].filter(Boolean) as ObjectProperty[],
							),
						),
					),
				);
				hasChanged = true;

				// Remove className/style JSXAttribute from component
				const newAttributes = component.openingElement.attributes.filter(
					(attr) => attr !== classNameProp && attr !== styleProp,
				);

				if (newAttributes.length !== component.openingElement.attributes.length) {
					hasChanged = true;
				}

				component.openingElement.attributes = newAttributes;
			}
		});
	});

	return hasChanged;
};

/**
 * Returns the file output while trying to avoid false positive codemod changes (e.g. unnecessary formatting changes).
 * - see: https://github.com/facebook/jscodeshift/issues/534
 * - see: https://github.com/facebook/jscodeshift/issues/262
 */
export const getSafeCodemoddedSource = ({
	root,
	fileInfo,
	hasChanged,
}: {
	root: ReturnType<API["jscodeshift"]>;
	fileInfo: FileInfo;
	hasChanged: boolean;
}) => {
	let lineTerminator;
	// If the file already has Windows line endings, then use Windows line endings
	if (/\r\n/.test(fileInfo.source)) {
		lineTerminator = "\r\n";
	} else {
		lineTerminator = "\n";
	}

	if (hasChanged) {
		// Return modified source code
		return root.toSource({
			lineTerminator: lineTerminator,
		});
	} else {
		// Return original source code
		return fileInfo.source;
	}
};
