import fs from "fs";
import path from "path";
import { SupportedCodemods } from "codemods/codemods";

export type LatestRunFile = {
	repoFolder: string;
	lastModified: number;
	files: Array<{
		filePath: string;
		originalFileContents: string;
		applicableCodemods: SupportedCodemods[];
		state: "active" | "finished-running" | "ignored";
		stages: Array<{
			codemod: SupportedCodemods;
			output: string;
		}>;
	}>;
};

/**
 * When this file is executed from a Next app, `__dirname` is somewhere inside `.next`
 * @returns true if `__dirname` is somewhere inside `.next`
 */
const isInNextFolder = (folder: string) => {
	return folder.split(path.sep).includes(".next");
};

const rootFolder = (() => {
	// If this file is executed from a Next app
	if (isInNextFolder(__dirname)) {
		let rootFolder = __dirname;
		// Keep going up the folder tree until we find the .next folder
		while (path.parse(rootFolder).name !== ".next") {
			rootFolder = path.join(rootFolder, "..");
		}
		// pop out of .next folder
		rootFolder = path.join(rootFolder, "..");
		return rootFolder;
	}
	// If this file is executed from somewhere else (E.g. unit tests)
	return path.join(__dirname, "..");
})();

export const getPreviousStageOutput = (
	file: LatestRunFile["files"][number],
	stageIndex: number,
): string => {
	// Return original file contents when stageIndex is 0
	if (stageIndex <= 0) {
		return file.originalFileContents;
	}

	// Return last stage when stageIndex is greater than the number of stages
	if (stageIndex > file.stages.length) {
		return file.stages[file.stages.length - 1].output;
	}

	return file.stages[stageIndex - 1].output;
};

export const getCurrentStageOutput = (
	file: LatestRunFile["files"][number],
	stageIndex: number,
): string | undefined => file?.stages[stageIndex]?.output;

/**
 * Adds or replaces a file in the latestRunFile
 * - If the file already exists, it will be replaced
 * - If the file does not exist, it will be added
 */
export const addToOrReplaceFileInLatestRunFile = (
	latestRunFile: LatestRunFile,
	filePath: string,
	fileToAdd: LatestRunFile["files"][number],
) => {
	const existingFileIndex = latestRunFile.files.findIndex((file) => file.filePath === filePath);
	if (existingFileIndex !== -1) {
		// Replace
		latestRunFile.files[existingFileIndex] = fileToAdd;
	} else {
		// Add
		latestRunFile.files.push(fileToAdd);
	}
};

export const writeLatestRunFile = (latestRunFile: LatestRunFile) => {
	latestRunFile.lastModified = Date.now();

	// Write to file
	fs.writeFileSync(
		path.join(rootFolder, "logs/v3", "latest_run.json"),
		JSON.stringify(latestRunFile, null, 2),
	);
};

/**
 * Rename the `latest_run.json`/`suggestions.json` to latest_run_<timestamp>.json/suggestions_<timestamp>.json (equivalent to deleting)
 */
export const deleteLatestRunFile = () => {
	const timestamp = Date.now();

	const newLatestRunFileName = `latest_run_${timestamp}.json`;
	const newSuggestionsFileName = `suggestions_${timestamp}.json`;

	fs.renameSync(
		path.join(rootFolder, "logs/v3", "latest_run.json"),
		path.join(rootFolder, "logs/v3", newLatestRunFileName),
	);
	fs.renameSync(
		path.join(rootFolder, "logs/v3", "suggestions.json"),
		path.join(rootFolder, "logs/v3", newSuggestionsFileName),
	);
};
