import { iTwinPackage, MatchLine, MatchRule } from "data/file_types";
import { Text } from "@itwin/itwinui-react";
import { MatchFile, ToastMessage } from "data/file_types";
import { getFolderInfo, getFolders, getMatchesInRepo } from "./api_calls";
import { globalToaster } from "modules/globals/toaster";

export const displayToast = (type: ToastMessage["type"], content: ToastMessage["content"]) => {
	if (globalToaster == null) {
		return;
	}

	globalToaster.setSettings({
		placement: "bottom-end",
		order: "descending",
	});
	if (type === "negative") {
		globalToaster.negative(content);
	} else if (type === "positive") {
		globalToaster.positive(content);
	} else {
		globalToaster.informational(content);
	}
};

export const getAllMatchesInRepo = async (
	repoFolder: string,
	setFoundRepoFolder: any,
	setMatches: any,
	setFileTree: any,
	setFolderInfo: any,
	gettingMatches: boolean,
	setGettingMatches: any,
	packagesToScan: iTwinPackage[],
) => {
	if (gettingMatches) {
		return;
	}
	if (repoFolder.trim().length === 0) {
		displayToast("negative", "Please enter repo path in the header's text input");
		return;
	}

	setGettingMatches(true);

	setMatches([]);
	setFoundRepoFolder(false);

	setFolderInfo(null);
	try {
		const folders = await getFolders(repoFolder);
		if (folders == null) {
			throw "Error: Try ensuring the folder path is valid";
		}

		const folderInfo = await getFolderInfo(repoFolder);
		if (folderInfo == null) {
			throw "Error: Try ensuring the folder path is valid";
		}
		setFolderInfo(folderInfo);
		setFoundRepoFolder(true);

		displayToast(
			"positive",
			`Repo folder found. Scanning ${folderInfo?.filesFoundCount} files for matches`,
		);

		setFileTree(folders);
		const matches = await getMatchesInRepo(gettingMatches, repoFolder, packagesToScan);
		if (matches == null) {
			throw "Error getting matches";
		}
		setMatches(matches);

		displayToast("positive", `Displaying found matches`);
	} catch (e: any) {
		displayToast("negative", e);
	}

	setGettingMatches(false);
};

/**
 * This function just modifies the passed `matches` and returns it.
 * You would have to manually update the state of `matches` (if needed).
 * @param matches
 * @param matchFileIndex
 * @param autoswapMatchLineIndices
 * @param shouldUndo
 * @returns
 */
export const getAutoswapChangesInFile = async (
	matches: MatchFile[],
	matchFileIndex: number,
	autoswapMatchLineIndices: number[],
	shouldUndo: boolean,
) => {
	const matchFile = matches[matchFileIndex];
	const response = await fetch(`/api/matches`, {
		method: "POST",
		headers: { "Content-Type": "application/json" },
		body: JSON.stringify({
			fileName: matchFile.fileName,
			lineMatches: [
				...matches[matchFileIndex].matchLines.filter((ml, index) =>
					autoswapMatchLineIndices.includes(index),
				),
			],
			shouldUndo,
		}),
	});
	const {
		successes,
		changedLines,
	}: {
		successes: boolean[];
		changedLines: boolean[];
	} = await response.json();

	for (let i = 0; i < autoswapMatchLineIndices.length; i++) {
		if (successes[i]) {
			matches[matchFileIndex].matchLines[autoswapMatchLineIndices[i]].changed = changedLines[i];
		}
	}

	if (successes.length < autoswapMatchLineIndices.length) {
		displayToast("negative", `Error ${shouldUndo ? "undoing change(s)" : "making change(s)"}`);
	}

	return matches;
};

export const implementAllAutoswapChanges = async (
	changed: boolean,
	matches: MatchFile[],
	setMatches: any,
	lockAllOperations: boolean,
	setLockAllOperations: any,
) => {
	if (lockAllOperations) {
		return;
	}
	setLockAllOperations(true);

	const newMatches = JSON.parse(JSON.stringify(matches));

	displayToast("informational", "Running batch task");
	for (let i = 0; i < newMatches.length; i++) {
		const autoswapLineIndices = [];

		for (let j = 0; j < newMatches[i].matchLines.length; j++) {
			if (newMatches[i].matchLines[j].type === "autoswap") {
				autoswapLineIndices.push(j);
			}
		}

		await getAutoswapChangesInFile(
			newMatches,
			i,
			autoswapLineIndices,
			!changed, // Since this parameter is shouldUndo. To make the change, shouldUndo should be false, and vice versa.
		);
	}
	displayToast(
		"positive",
		<Text>
			{changed
				? "Attempted all changes. Rescan to see if any changes could not be made."
				: "Attempted undoing all changes. Rescan to see if any changes could not be made."}
		</Text>,
	);

	setLockAllOperations(false);
	setMatches(newMatches);
};

export const toggleAllChangesInRule = async (
	matchRule: MatchRule,
	matchLines: (MatchLine & {
		matchFileIndex: number;
		matchLineIndex: number;
	})[],
	matches: MatchFile[],
	setMatches: any,
) => {
	const isAllChangedInRule = getIsAllChangedInRule(matchLines);
	const newMatches = JSON.parse(JSON.stringify(matches));

	displayToast(
		"informational",
		`Batch task: ${isAllChangedInRule ? "Undoing" : "Implementing"} all changes in rule...`,
	);

	if (matchRule.type === "autoswap") {
		for (const matchLine of matchLines) {
			await getAutoswapChangesInFile(
				newMatches,
				matchLine.matchFileIndex,
				[matchLine.matchLineIndex],
				isAllChangedInRule,
			);
		}
		setMatches(newMatches);
	} else {
		for (const matchLine of matchLines) {
			toggleNonAutoswapInFile(
				setMatches,
				matchLine.matchFileIndex,
				matchLine.matchLineIndex,
				!isAllChangedInRule,
			);
		}
	}

	displayToast(
		"positive",
		"All changes attempted. Rescan to see if any changes could not be made.",
	);
};

export const setAllManualChanges = (changed: boolean, matches: MatchFile[], setMatches: any) => {
	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
		for (let i = 0; i < newMatches.length; i++) {
			for (let j = 0; j < matches[i].matchLines.length; j++) {
				const matchLine = newMatches[i].matchLines[j];
				if (matchLine.type === "manual") {
					matchLine.changed = changed;
				}
			}
		}

		return newMatches;
	});
	displayToast(
		"positive",
		<Text>
			{`All "manual" changes marked as `}
			<strong>{changed ? `DONE` : `NOT DONE`}</strong>
		</Text>,
	);
};

export const setAllSuggestionChanges = (
	changed: boolean,
	matches: MatchFile[],
	setMatches: any,
) => {
	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
		for (let i = 0; i < newMatches.length; i++) {
			for (let j = 0; j < matches[i].matchLines.length; j++) {
				const matchLine = newMatches[i].matchLines[j];
				if (matchLine.type === "suggestion") {
					matchLine.changed = changed;
				}
			}
		}

		return newMatches;
	});
	displayToast(
		"positive",
		<Text>
			{`All "suggestion" changes marked as `}
			<strong>{changed ? `DONE` : `NOT DONE`}</strong>
		</Text>,
	);
};

export const getMatchFileIndex = (matches: MatchFile[], filePath: string) => {
	for (let i = 0; i < matches.length; i++) {
		if (matches[i].fileName === filePath) {
			return i;
		}
	}
	return -1;
};

export const getAllMatchLineCount = (matches: MatchFile[], matchLineType: MatchLine["type"]) => {
	let count = 0;

	for (const match of matches) {
		for (const matchLine of match.matchLines) {
			if (matchLine.type === matchLineType) {
				count++;
			}
		}
	}

	return count;
};

export const getAutoswapMatchIndices = (matches: MatchFile[], matchFileIndex: number): number[] => {
	const autoswapMatchLineIndices: number[] = [];

	for (let i = 0; i < matches[matchFileIndex].matchLines.length; i++) {
		if (matches[matchFileIndex].matchLines[i].type === "autoswap") {
			autoswapMatchLineIndices.push(i);
		}
	}
	return autoswapMatchLineIndices;
};

export const getNonAutoswapMatchIndices = (
	matches: MatchFile[],
	matchFileIndex: number,
): number[] => {
	const nonAutoswapMatchLineIndices: number[] = [];

	for (let i = 0; i < matches[matchFileIndex].matchLines.length; i++) {
		if (matches[matchFileIndex].matchLines[i].type !== "autoswap") {
			nonAutoswapMatchLineIndices.push(i);
		}
	}
	return nonAutoswapMatchLineIndices;
};

export const getManualMatchLineIndices = (matches: MatchFile[], matchFileIndex: number) => {
	const manualMatchLineIndices = [];

	for (let i = 0; i < matches[matchFileIndex].matchLines.length; i++) {
		if (matches[matchFileIndex].matchLines[i].type === "manual") {
			manualMatchLineIndices.push(i);
		}
	}

	return manualMatchLineIndices;
};

export const getSuggestionMatchLineIndices = (matches: MatchFile[], matchFileIndex: number) => {
	const suggestionMatchLineIndices = [];

	for (let i = 0; i < matches[matchFileIndex].matchLines.length; i++) {
		if (matches[matchFileIndex].matchLines[i].type === "suggestion") {
			suggestionMatchLineIndices.push(i);
		}
	}

	return suggestionMatchLineIndices;
};

export const getAutoswapMatchLines = (matches: MatchFile[], matchFileIndex: number) =>
	matches[matchFileIndex].matchLines.filter((l) => l.type === "autoswap");

export const getNonAutoswapMatchLines = (matches: MatchFile[], matchFileIndex: number) =>
	matches[matchFileIndex].matchLines.filter((l) => l.type !== "autoswap");

export const getManualMatchLines = (matches: MatchFile[], matchFileIndex: number) =>
	matches[matchFileIndex].matchLines.filter((l) => l.type === "manual");

export const getSuggestionMatchLines = (matches: MatchFile[], matchFileIndex: number) =>
	matches[matchFileIndex].matchLines.filter((l) => l.type === "suggestion");

export const getIsMatchLineMatchingFilter = (
	matchLine: MatchLine,
	selectedFilters: (MatchLine["type"] | "completed")[],
) => {
	return (
		selectedFilters.includes(matchLine.type) &&
		(!selectedFilters.includes("completed") || !matchLine.changed)
	);
};

export const getIsAllChangedCount = (matches: MatchFile[]) => {
	let count = 0;
	for (const matchFile of matches) {
		for (const matchLine of matchFile.matchLines) {
			if (matchLine.changed) {
				count++;
			}
		}
	}
	return count;
};

export const getIsAllChangedInRule = (
	matchLines: (MatchLine & {
		matchFileIndex: number;
		matchLineIndex: number;
	})[],
) => {
	return matchLines.filter((ml) => ml.changed).length === matchLines.length;
};

export const getIsAllChangedInFile = (matches: MatchFile[], matchFileIndex: number) => {
	const matchLines = matches[matchFileIndex].matchLines;
	return matchLines.filter((ml) => ml.changed).length === matchLines.length;
};

export const getIsAllAutoswapChangedInFile = (matches: MatchFile[], matchFileIndex: number) => {
	const autoswapMatchLines = getAutoswapMatchLines(matches, matchFileIndex);
	return autoswapMatchLines.filter((ml) => ml.changed).length === autoswapMatchLines.length;
};

export const getIsAllNonAutoswapChangedInFile = (matches: MatchFile[], matchFileIndex: number) => {
	const nonAutoswapMatchLines = getNonAutoswapMatchLines(matches, matchFileIndex);
	return nonAutoswapMatchLines.filter((ml) => ml.changed).length === nonAutoswapMatchLines.length;
};

export const getIsAllManualChangedInFile = (matches: MatchFile[], matchFileIndex: number) => {
	const manualMatchLines = getManualMatchLines(matches, matchFileIndex);
	return manualMatchLines.filter((ml) => ml.changed).length === manualMatchLines.length;
};

export const getIsAllSuggestionChangedInFile = (matches: MatchFile[], matchFileIndex: number) => {
	const suggestionMatchLines = getSuggestionMatchLines(matches, matchFileIndex);
	return suggestionMatchLines.filter((ml) => ml.changed).length === suggestionMatchLines.length;
};

export const toggleNonAutoswapInFile = (
	setMatches: any,
	matchFileIndex: number,
	matchLineIndex: number,
	changed: boolean,
) => {
	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
		newMatches[matchFileIndex].matchLines[matchLineIndex].changed = changed;
		return newMatches;
	});
};

export const toggleAllNonAutoswapInFile = (
	matches: MatchFile[],
	setMatches: any,
	matchFileIndex: number,
) => {
	const nonAutoswapMatchLineIndices = getNonAutoswapMatchIndices(matches, matchFileIndex);
	const isAllNonAutoswapChangedInFile = getIsAllNonAutoswapChangedInFile(matches, matchFileIndex);

	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));

		for (const index of nonAutoswapMatchLineIndices) {
			newMatches[matchFileIndex].matchLines[index].changed = !isAllNonAutoswapChangedInFile;
		}
		return newMatches;
	});
};

export const toggleAllManualInFile = (
	matches: MatchFile[],
	setMatches: any,
	matchFileIndex: number,
) => {
	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
		const manualMatchLineIndices: number[] = getManualMatchLineIndices(matches, matchFileIndex);
		const isAllManualChangedInFile = getIsAllManualChangedInFile(matches, matchFileIndex);

		for (const index of manualMatchLineIndices) {
			newMatches[matchFileIndex].matchLines[index].changed = !isAllManualChangedInFile;
		}
		return newMatches;
	});
};

export const toggleAllSuggestionInFile = (
	matches: MatchFile[],
	setMatches: any,
	matchFileIndex: number,
) => {
	setMatches((prev: MatchFile[]) => {
		const newMatches: MatchFile[] = JSON.parse(JSON.stringify(prev));
		const suggestionMatchLineIndices: number[] = getSuggestionMatchLineIndices(
			matches,
			matchFileIndex,
		);
		const isAllSuggestionChangedInFile = getIsAllSuggestionChangedInFile(matches, matchFileIndex);

		for (const index of suggestionMatchLineIndices) {
			newMatches[matchFileIndex].matchLines[index].changed = !isAllSuggestionChangedInFile;
		}
		return newMatches;
	});
};
